# Copyright (c) 2019 Uber Technologies, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
from typing import Callable

import pytest
import torch

from ludwig.utils.image_utils import (
    crop,
    crop_or_pad,
    grayscale,
    num_channels_in_image,
    pad,
    resize_image,
    ResizeChannels,
)


@pytest.mark.parametrize("pad_fn", [pad, torch.jit.script(pad)])
@pytest.mark.parametrize(
    "img,size,padded_img",
    [
        (
            torch.arange(12, dtype=torch.int).reshape(3, 2, 2),
            4,
            torch.Tensor(
                [
                    0,
                    0,
                    1,
                    1,
                    0,
                    0,
                    1,
                    1,
                    2,
                    2,
                    3,
                    3,
                    2,
                    2,
                    3,
                    3,
                    4,
                    4,
                    5,
                    5,
                    4,
                    4,
                    5,
                    5,
                    6,
                    6,
                    7,
                    7,
                    6,
                    6,
                    7,
                    7,
                    8,
                    8,
                    9,
                    9,
                    8,
                    8,
                    9,
                    9,
                    10,
                    10,
                    11,
                    11,
                    10,
                    10,
                    11,
                    11,
                ]
            )
            .type(torch.int)
            .reshape(3, 4, 4),
        )
    ],
)
def test_pad(pad_fn: Callable, img: torch.Tensor, size: int, padded_img: torch.Tensor):
    output_img = pad_fn(img, size)
    assert torch.equal(output_img, padded_img)


@pytest.mark.parametrize("crop_fn", [crop, torch.jit.script(crop)])
@pytest.mark.parametrize(
    "img,size,cropped_img",
    [
        (
            torch.arange(27, dtype=torch.int).reshape(3, 3, 3),
            2,
            torch.Tensor([0, 1, 3, 4, 9, 10, 12, 13, 18, 19, 21, 22]).type(torch.int).reshape(3, 2, 2),
        )
    ],
)
def test_crop(crop_fn: Callable, img: torch.Tensor, size: int, cropped_img: torch.Tensor):
    output_img = crop_fn(img, size)
    assert torch.equal(output_img, cropped_img)


@pytest.mark.parametrize("crop_or_pad_fn", [crop_or_pad, torch.jit.script(crop_or_pad)])
@pytest.mark.parametrize(
    "img,new_size,expected_img",
    [
        (
            torch.arange(12, dtype=torch.int).reshape(3, 2, 2),
            4,
            torch.Tensor(
                [
                    0,
                    0,
                    1,
                    1,
                    0,
                    0,
                    1,
                    1,
                    2,
                    2,
                    3,
                    3,
                    2,
                    2,
                    3,
                    3,
                    4,
                    4,
                    5,
                    5,
                    4,
                    4,
                    5,
                    5,
                    6,
                    6,
                    7,
                    7,
                    6,
                    6,
                    7,
                    7,
                    8,
                    8,
                    9,
                    9,
                    8,
                    8,
                    9,
                    9,
                    10,
                    10,
                    11,
                    11,
                    10,
                    10,
                    11,
                    11,
                ]
            )
            .type(torch.int)
            .reshape(3, 4, 4),
        ),
        (
            torch.arange(27, dtype=torch.int).reshape(3, 3, 3),
            2,
            torch.Tensor([0, 1, 3, 4, 9, 10, 12, 13, 18, 19, 21, 22]).type(torch.int).reshape(3, 2, 2),
        ),
    ],
)
def test_crop_or_pad(crop_or_pad_fn: Callable, img: torch.Tensor, new_size: int, expected_img: torch.Tensor):
    output_image = crop_or_pad_fn(img, new_size)
    assert torch.equal(output_image, expected_img)


@pytest.mark.parametrize("resize_image_fn", [resize_image, torch.jit.script(resize_image)])
@pytest.mark.parametrize(
    "img,new_size,resize_method,expected_img",
    [
        (
            torch.arange(27, dtype=torch.int).reshape(3, 3, 3),
            2,
            "crop_or_pad",
            torch.Tensor([0, 1, 3, 4, 9, 10, 12, 13, 18, 19, 21, 22]).type(torch.int).reshape(3, 2, 2),
        ),
        (
            torch.arange(27, dtype=torch.int).reshape(3, 3, 3),
            2,
            "interpolate",
            torch.Tensor([1, 2, 6, 7, 10, 12, 14, 16, 19, 20, 24, 25]).type(torch.int).reshape(3, 2, 2),
        ),
    ],
)
def test_resize_image(
    resize_image_fn: Callable, img: torch.Tensor, new_size: int, resize_method: str, expected_img: torch.Tensor
):
    output_img = resize_image_fn(img, new_size, resize_method)
    assert torch.equal(output_img, expected_img)


@pytest.mark.parametrize("grayscale_fn", [grayscale, torch.jit.script(grayscale)])
@pytest.mark.parametrize(
    "input_img,grayscale_img",
    [(torch.arange(12).reshape(3, 2, 2).type(torch.int), torch.Tensor([[[3, 4], [5, 6]]]).type(torch.int))],
)
def test_grayscale(grayscale_fn: Callable, input_img: torch.Tensor, grayscale_img: torch.Tensor):
    output_img = grayscale_fn(input_img)
    assert torch.equal(output_img, grayscale_img)


def test_num_channels_in_image():
    image_2d = torch.randint(0, 1, (10, 10))
    image_3d = torch.randint(0, 1, (3, 10, 10))
    assert num_channels_in_image(image_2d) == 1
    assert num_channels_in_image(image_3d) == 3

    with pytest.raises(ValueError):
        num_channels_in_image(torch.rand(5))
        num_channels_in_image(None)


@pytest.mark.parametrize("image_shape", [(1, 10, 10), (3, 10, 10), (5, 10, 10)])
@pytest.mark.parametrize("num_channels_expected", [1, 2, 3, 4])
def test_ResizeChannels_module(image_shape, num_channels_expected):
    image = torch.randint(0, 1, image_shape)
    fn = ResizeChannels(num_channels_expected)
    assert fn(image).shape == tuple([num_channels_expected] + list(image_shape[1:]))


@pytest.mark.parametrize("image_shape", [(2, 1, 10, 10), (2, 3, 10, 10), (2, 5, 10, 10)])
@pytest.mark.parametrize("num_channels_expected", [1, 2, 3, 4])
def test_ResizeChannels_module_with_batch_dim(image_shape, num_channels_expected):
    image = torch.randint(0, 1, image_shape)
    fn = ResizeChannels(num_channels_expected)
    assert fn(image).shape == tuple([image_shape[0], num_channels_expected] + list(image_shape[2:]))



import pandas as pd

from ludwig.backend import LocalBackend
from ludwig.features import text_feature


def test_backwards_compatibility():
    # Tests that legacy level-based metadata keys are supported.
    metadata = {
        "SibSp": {
            "char_idx2str": ["<EOS>", "<SOS>", "<PAD>", "<UNK>", "0", "1", "2", "4", "3", "8", "5"],
            "char_max_sequence_length": 3,
            "char_pad_idx": 2,
            "char_pad_symbol": "<PAD>",
            "char_str2freq": {
                "0": 608,
                "1": 209,
                "2": 28,
                "3": 16,
                "4": 18,
                "5": 5,
                "8": 7,
                "<EOS>": 0,
                "<PAD>": 0,
                "<SOS>": 0,
                "<UNK>": 0,
            },
            "char_str2idx": {
                "0": 4,
                "1": 5,
                "2": 6,
                "3": 8,
                "4": 7,
                "5": 10,
                "8": 9,
                "<EOS>": 0,
                "<PAD>": 2,
                "<SOS>": 1,
                "<UNK>": 3,
            },
            "char_unk_symbol": "<UNK>",
            "char_vocab_size": 11,
            "preprocessing": {
                "char_most_common": 70,
                "char_sequence_length_limit": 1024,
                "char_tokenizer": "characters",
                "char_vocab_file": None,
                "computed_fill_value": "<UNK>",
                "fill_value": "<UNK>",
                "lowercase": True,
                "missing_value_strategy": "fill_with_const",
                "padding": "right",
                "padding_symbol": "<PAD>",
                "pretrained_model_name_or_path": None,
                "unknown_symbol": "<UNK>",
                "word_most_common": 20000,
                "word_sequence_length_limit": 256,
                "word_tokenizer": "space_punct",
                "word_vocab_file": None,
            },
            "word_idx2str": ["<EOS>", "<SOS>", "<PAD>", "<UNK>", "0", "1", "2", "4", "3", "8", "5"],
            "word_max_sequence_length": 3,
            "word_pad_idx": 2,
            "word_pad_symbol": "<PAD>",
            "word_str2freq": {
                "0": 608,
                "1": 209,
                "2": 28,
                "3": 16,
                "4": 18,
                "5": 5,
                "8": 7,
                "<EOS>": 0,
                "<PAD>": 0,
                "<SOS>": 0,
                "<UNK>": 0,
            },
            "word_str2idx": {
                "0": 4,
                "1": 5,
                "2": 6,
                "3": 8,
                "4": 7,
                "5": 10,
                "8": 9,
                "<EOS>": 0,
                "<PAD>": 2,
                "<SOS>": 1,
                "<UNK>": 3,
            },
            "word_unk_symbol": "<UNK>",
            "word_vocab_size": 11,
        }
    }

    column = pd.core.series.Series(["hello world", "hello world"])

    feature_data = text_feature.TextInputFeature.feature_data(
        column, metadata["SibSp"], metadata["SibSp"]["preprocessing"], LocalBackend()
    )

    assert list(feature_data[0]) == [1, 3, 3]
    assert list(feature_data[1]) == [1, 3, 3]



import csv
import logging
import os
from dataclasses import dataclass
from statistics import mean
from typing import Dict, List, Optional, Set, Union

import ludwig.modules.metric_modules  # noqa: F401
from ludwig.benchmarking.utils import format_memory, format_time
from ludwig.globals import MODEL_HYPERPARAMETERS_FILE_NAME
from ludwig.modules.metric_registry import get_metric_classes, metric_feature_type_registry  # noqa: F401
from ludwig.types import ModelConfigDict
from ludwig.utils.data_utils import load_json

logger = logging.getLogger()


@dataclass
class MetricDiff:
    """Diffs for a metric."""

    # Name of the metric.
    name: str

    # Value of the metric in base experiment (the one we benchmark against).
    base_value: float

    # Value of the metric in the experimental experiment.
    experimental_value: float

    # experimental_value - base_value.
    diff: float

    # Percentage of change the metric with respect to base_value.
    diff_percentage: Union[float, str]

    def __post_init__(self):
        """Add human-readable string representations to the field."""

        if "memory" in self.name:
            self.base_value_str = format_memory(self.base_value)
            self.experimental_value_str = format_memory(self.experimental_value)
            self.diff_str = format_memory(self.diff)
        elif "time" in self.name:
            self.base_value_str = format_time(self.base_value)
            self.experimental_value_str = format_time(self.experimental_value)
            self.diff_str = format_time(self.diff)
        else:
            self.base_value_str = str(self.base_value)
            self.experimental_value_str = str(self.experimental_value)
            self.diff_str = str(self.diff)


def build_diff(name: str, base_value: float, experimental_value: float) -> MetricDiff:
    """Build a diff between any type of metric.

    :param name: name assigned to the metric to be diff-ed.
    :param base_value: base value of the metric.
    :param experimental_value: experimental value of the metric.
    """
    diff = experimental_value - base_value
    diff_percentage = 100 * diff / base_value if base_value != 0 else "inf"

    return MetricDiff(
        name=name,
        base_value=base_value,
        experimental_value=experimental_value,
        diff=diff,
        diff_percentage=diff_percentage,
    )


##############################
# Resource Usage Dataclasses #
##############################


@dataclass
class MetricsSummary:
    """Summary of metrics from one experiment."""

    # Path containing the artifacts for the experiment.
    experiment_local_directory: str

    # Full Ludwig config.
    config: ModelConfigDict

    # LudwigModel output feature type.
    output_feature_type: str

    # LudwigModel output feature name.
    output_feature_name: str

    # Dictionary that maps from metric name to their values.
    metric_to_values: Dict[str, Union[float, int]]

    # Names of metrics for the output feature.
    metric_names: Set[str]


@dataclass
class MetricsDiff:
    """Store diffs for two experiments."""

    # Dataset the two experiments are being compared on.
    dataset_name: str

    # Name of the base experiment (the one we benchmark against).
    base_experiment_name: str

    # Name of the experimental experiment.
    experimental_experiment_name: str

    # Path under which all artifacts live on the local machine.
    local_directory: str

    # `MetricsSummary` of the base_experiment.
    base_summary: MetricsSummary

    # `MetricsSummary` of the experimental_experiment.
    experimental_summary: MetricsSummary

    # `List[MetricDiff]` containing diffs for metric of the two experiments.
    metrics: List[MetricDiff]

    def to_string(self):
        ret = []
        spacing_str = "{:<20} {:<33} {:<13} {:<13} {:<13} {:<5}"
        ret.append(
            spacing_str.format(
                "Output Feature Name",
                "Metric Name",
                self.base_experiment_name,
                self.experimental_experiment_name,
                "Diff",
                "Diff Percentage",
            )
        )
        for metric in sorted(self.metrics, key=lambda m: m.name):
            output_feature_name = self.base_summary.output_feature_name
            metric_name = metric.name
            experiment1_val = round(metric.base_value, 3)
            experiment2_val = round(metric.experimental_value, 3)
            diff = round(metric.diff, 3)
            diff_percentage = metric.diff_percentage
            if isinstance(diff_percentage, float):
                diff_percentage = round(metric.diff_percentage, 3)
            ret.append(
                spacing_str.format(
                    output_feature_name,
                    metric_name,
                    experiment1_val,
                    experiment2_val,
                    diff,
                    diff_percentage,
                )
            )
        return "\n".join(ret)


def export_metrics_diff_to_csv(metrics_diff: MetricsDiff, path: str):
    """Export metrics report to .csv.

    :param metrics_diff: MetricsDiff object containing the diff for two experiments on a dataset.
    :param path: file name of the exported csv.
    """
    with open(path, "w", newline="") as f:
        writer = csv.DictWriter(
            f,
            fieldnames=[
                "Dataset Name",
                "Output Feature Name",
                "Metric Name",
                metrics_diff.base_experiment_name,
                metrics_diff.experimental_experiment_name,
                "Diff",
                "Diff Percentage",
            ],
        )
        writer.writeheader()

        for metric in sorted(metrics_diff.metrics, key=lambda m: m.name):
            output_feature_name = metrics_diff.base_summary.output_feature_name
            metric_name = metric.name
            experiment1_val = round(metric.base_value, 3)
            experiment2_val = round(metric.experimental_value, 3)
            diff = round(metric.diff, 3)
            diff_percentage = metric.diff_percentage
            if isinstance(diff_percentage, float):
                diff_percentage = round(metric.diff_percentage, 3)
            writer.writerow(
                {
                    "Dataset Name": metrics_diff.dataset_name,
                    "Output Feature Name": output_feature_name,
                    "Metric Name": metric_name,
                    metrics_diff.base_experiment_name: experiment1_val,
                    metrics_diff.experimental_experiment_name: experiment2_val,
                    "Diff": diff,
                    "Diff Percentage": diff_percentage,
                }
            )
        logger.info(f"Exported a CSV report to {path}\n")


def build_metrics_summary(experiment_local_directory: str) -> MetricsSummary:
    """Build a metrics summary for an experiment.

    :param experiment_local_directory: directory where the experiment artifacts live.
        e.g. local_experiment_repo/ames_housing/some_experiment/
    """
    config = load_json(
        os.path.join(experiment_local_directory, "experiment_run", "model", MODEL_HYPERPARAMETERS_FILE_NAME)
    )
    report = load_json(os.path.join(experiment_local_directory, "experiment_run", "test_statistics.json"))
    output_feature_type: str = config["output_features"][0]["type"]
    output_feature_name: str = config["output_features"][0]["name"]
    metric_dict = report[output_feature_name]
    full_metric_names = get_metric_classes(output_feature_type)
    metric_to_values: Dict[str, Union[float, int]] = {
        metric_name: metric_dict[metric_name] for metric_name in full_metric_names if metric_name in metric_dict
    }
    metric_names: Set[str] = set(metric_to_values)

    return MetricsSummary(
        experiment_local_directory=experiment_local_directory,
        config=config,
        output_feature_name=output_feature_name,
        output_feature_type=output_feature_type,
        metric_to_values=metric_to_values,
        metric_names=metric_names,
    )


def build_metrics_diff(
    dataset_name: str, base_experiment_name: str, experimental_experiment_name: str, local_directory: str
) -> MetricsDiff:
    """Build a MetricsDiff object between two experiments on a dataset.

    :param dataset_name: the name of the Ludwig dataset.
    :param base_experiment_name: the name of the base experiment.
    :param experimental_experiment_name: the name of the experimental experiment.
    :param local_directory: the local directory where the experiment artifacts are downloaded.
    """
    base_summary: MetricsSummary = build_metrics_summary(
        os.path.join(local_directory, dataset_name, base_experiment_name)
    )
    experimental_summary: MetricsSummary = build_metrics_summary(
        os.path.join(local_directory, dataset_name, experimental_experiment_name)
    )

    metrics_in_common = set(base_summary.metric_names).intersection(set(experimental_summary.metric_names))

    metrics: List[MetricDiff] = [
        build_diff(name, base_summary.metric_to_values[name], experimental_summary.metric_to_values[name])
        for name in metrics_in_common
    ]

    return MetricsDiff(
        dataset_name=dataset_name,
        base_experiment_name=base_experiment_name,
        experimental_experiment_name=experimental_experiment_name,
        local_directory=local_directory,
        base_summary=base_summary,
        experimental_summary=experimental_summary,
        metrics=metrics,
    )


##############################
# Resource Usage Dataclasses #
##############################


@dataclass
class ResourceUsageSummary:
    """Summary of resource usage metrics from one experiment."""

    # The tag with which the code block/function is labeled.
    code_block_tag: str

    # Dictionary that maps from metric name to their values.
    metric_to_values: Dict[str, Union[float, int]]

    # Names of metrics for the output feature.
    metric_names: Set[str]


@dataclass
class ResourceUsageDiff:
    """Store resource usage diffs for two experiments."""

    # The tag with which the code block/function is labeled.
    code_block_tag: str

    # Name of the base experiment (the one we benchmark against).
    base_experiment_name: str

    # Name of the experimental experiment.
    experimental_experiment_name: str

    # `List[Diff]` containing diffs for metric of the two experiments.
    metrics: List[MetricDiff]

    def to_string(self):
        ret = []
        spacing_str = "{:<36} {:<20} {:<20} {:<20} {:<5}"
        ret.append(
            spacing_str.format(
                "Metric Name",
                self.base_experiment_name,
                self.experimental_experiment_name,
                "Diff",
                "Diff Percentage",
            )
        )
        for metric in sorted(self.metrics, key=lambda m: m.name):
            diff_percentage = metric.diff_percentage
            if isinstance(metric.diff_percentage, float):
                diff_percentage = round(metric.diff_percentage, 3)
            ret.append(
                spacing_str.format(
                    metric.name,
                    metric.base_value_str,
                    metric.experimental_value_str,
                    metric.diff_str,
                    diff_percentage,
                )
            )
        return "\n".join(ret)


def export_resource_usage_diff_to_csv(resource_usage_diff: ResourceUsageDiff, path: str):
    """Export resource usage metrics report to .csv.

    :param resource_usage_diff: ResourceUsageDiff object containing the diff for two experiments on a dataset.
    :param path: file name of the exported csv.
    """
    with open(path, "w", newline="") as f:
        writer = csv.DictWriter(
            f,
            fieldnames=[
                "Code Block Tag",
                "Metric Name",
                resource_usage_diff.base_experiment_name,
                resource_usage_diff.experimental_experiment_name,
                "Diff",
                "Diff Percentage",
            ],
        )
        writer.writeheader()

        for metric in sorted(resource_usage_diff.metrics, key=lambda m: m.name):
            diff_percentage = metric.diff_percentage
            if isinstance(metric.diff_percentage, float):
                diff_percentage = round(metric.diff_percentage, 3)
            writer.writerow(
                {
                    "Code Block Tag": resource_usage_diff.code_block_tag,
                    "Metric Name": metric.name,
                    resource_usage_diff.base_experiment_name: metric.base_value_str,
                    resource_usage_diff.experimental_experiment_name: metric.experimental_value_str,
                    "Diff": metric.diff_str,
                    "Diff Percentage": diff_percentage,
                }
            )
        logger.info(f"Exported a CSV report to {path}\n")


def average_runs(path_to_runs_dir: str) -> Dict[str, Union[int, float]]:
    """Return average metrics from code blocks/function that ran more than once.

    Metrics for code blocks/functions that were executed exactly once will be returned as is.

    :param path_to_runs_dir: path to where metrics specific to a tag are stored.
        e.g. resource_usage_out_dir/torch_ops_resource_usage/LudwigModel.evaluate/
        This directory will contain JSON files with the following pattern run_*.json
    """
    runs = [load_json(os.path.join(path_to_runs_dir, run)) for run in os.listdir(path_to_runs_dir)]
    # asserting that keys to each of the dictionaries are consistent throughout the runs.
    assert len(runs) == 1 or all(runs[i].keys() == runs[i + 1].keys() for i in range(len(runs) - 1))
    runs_average = {"num_runs": len(runs)}
    for key in runs[0]:
        if isinstance(runs[0][key], (int, float)):
            runs_average[key] = mean([run[key] for run in runs])
    return runs_average


def summarize_resource_usage(path: str, tags: Optional[List[str]] = None) -> List[ResourceUsageSummary]:
    """Create resource usage summaries for each code block/function that was decorated with ResourceUsageTracker.

    Each entry of the list corresponds to the metrics collected from a code block/function run.
    Important: code blocks that ran more than once are averaged.

    :param path: corresponds to the `output_dir` argument in a ResourceUsageTracker run.
    :param tags: (optional) list of tags to create summary for. If None, metrics from all tags will be summarized.
    """
    summary = dict()
    # metric types: system_resource_usage, torch_ops_resource_usage.
    all_metric_types = {"system_resource_usage", "torch_ops_resource_usage"}
    for metric_type in all_metric_types.intersection(os.listdir(path)):
        metric_type_path = os.path.join(path, metric_type)
        # code block tags correspond to the `tag` argument in ResourceUsageTracker.
        for code_block_tag in os.listdir(metric_type_path):
            if tags and code_block_tag not in tags:
                continue
            if code_block_tag not in summary:
                summary[code_block_tag] = {}
            run_path = os.path.join(metric_type_path, code_block_tag)
            # Metrics from code blocks/functions that ran more than once are averaged.
            summary[code_block_tag][metric_type] = average_runs(run_path)

    summary_list = []
    for code_block_tag, metric_type_dicts in summary.items():
        merged_summary: Dict[str, Union[float, int]] = {}
        for metrics in metric_type_dicts.values():
            assert "num_runs" in metrics
            assert "num_runs" not in merged_summary or metrics["num_runs"] == merged_summary["num_runs"]
            merged_summary.update(metrics)
        summary_list.append(
            ResourceUsageSummary(
                code_block_tag=code_block_tag, metric_to_values=merged_summary, metric_names=set(merged_summary)
            )
        )
    return summary_list


def build_resource_usage_diff(
    base_path: str,
    experimental_path: str,
    base_experiment_name: Optional[str] = None,
    experimental_experiment_name: Optional[str] = None,
) -> List[ResourceUsageDiff]:
    """Build and return a ResourceUsageDiff object to diff resource usage metrics between two experiments.

    :param base_path: corresponds to the `output_dir` argument in the base ResourceUsageTracker run.
    :param experimental_path: corresponds to the `output_dir` argument in the experimental ResourceUsageTracker run.
    """
    base_summary_list = summarize_resource_usage(base_path)
    experimental_summary_list = summarize_resource_usage(experimental_path)

    summaries_list = []
    for base_summary in base_summary_list:
        for experimental_summary in experimental_summary_list:
            if base_summary.code_block_tag == experimental_summary.code_block_tag:
                summaries_list.append((base_summary, experimental_summary))

    diffs = []
    for base_summary, experimental_summary in summaries_list:
        metrics_in_common = set(base_summary.metric_names).intersection(set(experimental_summary.metric_names))
        metrics: List[MetricDiff] = [
            build_diff(name, base_summary.metric_to_values[name], experimental_summary.metric_to_values[name])
            for name in metrics_in_common
        ]
        diff = ResourceUsageDiff(
            code_block_tag=base_summary.code_block_tag,
            base_experiment_name=base_experiment_name if base_experiment_name else "experiment_1",
            experimental_experiment_name=experimental_experiment_name
            if experimental_experiment_name
            else "experiment_2",
            metrics=metrics,
        )
        diffs.append(diff)
    return diffs



# Copyright (c) 2022 Predibase, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

from ludwig.api_annotations import PublicAPI


@PublicAPI
class LudwigError(Exception):
    """Base class for all custom exceptions raised by the Ludwig framework."""

    def __reduce__(self):
        """Docs: https://docs.python.org/3/library/pickle.html#object.__reduce__."""
        raise NotImplementedError(
            "Implement __reduce__ for all subclasses of LudwigError as it's necessary for "
            "serialization by Ray. See https://github.com/ludwig-ai/ludwig/pull/2695."
        )


@PublicAPI
class InputDataError(LudwigError, ValueError):
    """Exception raised for errors in the input data.

    Appropriate for data which is not convertible to the input feature type, columns with all missing values,
    categorical columns with only one category, etc...

    Attributes:
        column - The name of the input column which caused the error
        feature_type - The Ludwig feature type which caused the error (number, binary, category...).
        message - An error message describing the situation.
    """

    def __init__(self, column_name: str, feature_type: str, message: str):
        self.column_name = column_name
        self.feature_type = feature_type
        self.message = message
        super().__init__(message)

    def __str__(self):
        return f'Column "{self.column_name}" as {self.feature_type} feature: {self.message}'

    def __reduce__(self):
        return type(self), (self.column_name, self.feature_type, self.message)


@PublicAPI
class ConfigValidationError(LudwigError, ValueError):
    """Exception raised for errors in the Ludwig configuration.

    Appropriate for bad configuration values, missing required configuration values, etc...

    Attributes:
        message - An error message describing the situation.
    """

    def __init__(self, message: str):
        self.message = message
        super().__init__(message)

    def __reduce__(self):
        return type(self), (self.message,)



from typing import List

from ludwig.api_annotations import DeveloperAPI
from ludwig.constants import IMAGE, MODEL_ECD
from ludwig.schema.encoders.base import BaseEncoderConfig
from ludwig.schema.encoders.utils import EncoderDataclassField
from ludwig.schema.features.augmentation.base import BaseAugmentationConfig
from ludwig.schema.features.augmentation.image import RandomHorizontalFlipConfig, RandomRotateConfig
from ludwig.schema.features.augmentation.utils import AugmentationDataclassField
from ludwig.schema.features.base import BaseInputFeatureConfig
from ludwig.schema.features.preprocessing.base import BasePreprocessingConfig
from ludwig.schema.features.preprocessing.utils import PreprocessingDataclassField
from ludwig.schema.features.utils import defaults_config_registry, ecd_input_config_registry, input_mixin_registry
from ludwig.schema.utils import BaseMarshmallowConfig, ludwig_dataclass

# Augmentation operations when augmentation is set to True
AUGMENTATION_DEFAULT_OPERATIONS = [
    RandomHorizontalFlipConfig(),
    RandomRotateConfig(),
]


@DeveloperAPI
@defaults_config_registry.register(IMAGE)
@input_mixin_registry.register(IMAGE)
@ludwig_dataclass
class ImageInputFeatureConfigMixin(BaseMarshmallowConfig):
    """ImageInputFeatureConfigMixin is a dataclass that configures the parameters used in both the image input
    feature and the image global defaults section of the Ludwig Config."""

    preprocessing: BasePreprocessingConfig = PreprocessingDataclassField(feature_type=IMAGE)

    encoder: BaseEncoderConfig = EncoderDataclassField(
        MODEL_ECD,
        feature_type=IMAGE,
        default="stacked_cnn",
    )

    augmentation: List[BaseAugmentationConfig] = AugmentationDataclassField(
        feature_type=IMAGE,
        default=False,
        default_augmentations=AUGMENTATION_DEFAULT_OPERATIONS,
        description="Augmentation operation configuration.",
    )

    def has_augmentation(self) -> bool:
        # Check for None, False, and []
        return bool(self.augmentation)


@DeveloperAPI
@ecd_input_config_registry.register(IMAGE)
@ludwig_dataclass
class ImageInputFeatureConfig(ImageInputFeatureConfigMixin, BaseInputFeatureConfig):
    """ImageInputFeatureConfig is a dataclass that configures the parameters used for an image input feature."""

    pass



import contextlib
import os
import shutil

import numpy as np
import pandas as pd
import pytest

from ludwig.api import LudwigModel
from ludwig.backend import LocalBackend
from tests.integration_tests.utils import create_data_set_to_use, RAY_BACKEND_CONFIG, spawn

try:
    import ray

    from ludwig.backend.ray import RayBackend
except ImportError:
    ray = None

rs = np.random.RandomState(42)


@contextlib.contextmanager
def ray_start(num_cpus=2, num_gpus=None):
    res = ray.init(
        num_cpus=num_cpus,
        num_gpus=num_gpus,
        include_dashboard=False,
        object_store_memory=150 * 1024 * 1024,
    )
    try:
        yield res
    finally:
        ray.shutdown()


@spawn
def run_test_imbalance_ray(
    tmpdir,
    input_df,
    config,
    balance,
    num_cpus=2,
    num_gpus=None,
):
    with ray_start(num_cpus=num_cpus, num_gpus=num_gpus):
        csv_filename = os.path.join(tmpdir, "dataset.csv")
        input_df.to_csv(csv_filename)
        dataset_parquet = create_data_set_to_use("parquet", csv_filename)

        model = LudwigModel(config, backend=RAY_BACKEND_CONFIG, callbacks=None)
        output_dir = None

        try:
            _, output_dataset, output_dir = model.train(
                dataset=dataset_parquet,
                training_set=None,
                validation_set=None,
                test_set=None,
                skip_save_processed_input=True,
                skip_save_progress=True,
                skip_save_unprocessed_output=True,
                skip_save_log=True,
            )
        finally:
            # Remove results/intermediate data saved to disk
            shutil.rmtree(output_dir, ignore_errors=True)

        input_train_set = input_df.sample(frac=0.7, replace=False)
        processed_len = output_dataset[0].ds.count()
        processed_target_pos = output_dataset[0].ds.sum(on="Label_mZFLky")
        processed_target_neg = output_dataset[0].ds.count() - output_dataset[0].ds.sum(on="Label_mZFLky")
        assert len(input_train_set) == 140
        assert 0.05 <= len(input_train_set[input_train_set["Label"] == 1]) / len(input_train_set) <= 0.15
        assert round(processed_target_pos / processed_target_neg, 1) == 0.5
        assert model.backend.df_engine.parallelism == RAY_BACKEND_CONFIG["processor"]["parallelism"]
        assert isinstance(model.backend, RayBackend)

        if balance == "oversample_minority":
            assert len(input_train_set) < processed_len

        if balance == "undersample_majority":
            assert len(input_train_set) > processed_len


def run_test_imbalance_local(
    input_df,
    config,
    balance,
):
    model = LudwigModel(config)
    _, output_dataset, output_dir = model.train(
        input_df,
        skip_save_model=True,
        skip_save_log=True,
        skip_save_progress=True,
        skip_save_processed_input=True,
        skip_save_training_description=True,
        skip_save_training_statistics=True,
    )

    input_train_set = input_df.sample(frac=0.7, replace=False)
    processed_len = output_dataset[0].size
    processed_target_pos = sum(output_dataset[0].dataset["Label_2Xl8CP"])
    processed_target_neg = len(output_dataset[0].dataset["Label_2Xl8CP"]) - processed_target_pos
    assert len(input_train_set) == 140
    assert 0.05 <= len(input_train_set[input_train_set["Label"] == 1]) / len(input_train_set) <= 0.15
    assert round(processed_target_pos / processed_target_neg, 1) == 0.5
    assert isinstance(model.backend, LocalBackend)

    if balance == "oversample_minority":
        assert len(input_train_set) < processed_len
        assert 55 <= processed_target_pos <= 75
        assert 110 <= processed_target_neg <= 150

    if balance == "undersample_majority":
        assert len(input_train_set) > processed_len
        assert 7 <= processed_target_pos <= 20
        assert 14 <= processed_target_neg <= 40


@pytest.mark.parametrize(
    "balance",
    ["oversample_minority", "undersample_majority"],
)
@pytest.mark.distributed
@pytest.mark.skip(reason="Flaky")
def test_imbalance_ray(balance):
    config = {
        "input_features": [
            {"name": "Index", "column": "Index", "type": "numerical"},
            {"name": "random_1", "column": "random_1", "type": "numerical"},
            {"name": "random_2", "column": "random_2", "type": "numerical"},
        ],
        "output_features": [{"name": "Label", "column": "Label", "type": "binary"}],
        "trainer": {"epochs": 2, "batch_size": 8},
        "preprocessing": {},
    }
    split_col = np.concatenate((np.zeros(140), np.ones(20), np.full(40, 2)))
    rs.shuffle(split_col)
    df = pd.DataFrame(
        {
            "Index": np.arange(0, 200, 1),
            "random_1": np.random.randint(0, 50, 200),
            "random_2": np.random.choice(["Type A", "Type B", "Type C", "Type D"], 200),
            "Label": np.concatenate((np.zeros(180), np.ones(20))),
            "split": split_col,
        }
    )

    config["preprocessing"][balance] = 0.5
    run_test_imbalance_ray(df, config, balance)


@pytest.mark.parametrize(
    "balance",
    ["oversample_minority", "undersample_majority"],
)
def test_imbalance_local(balance):
    config = {
        "input_features": [
            {"name": "Index", "column": "Index", "type": "number"},
            {"name": "random_1", "column": "random_1", "type": "number"},
            {"name": "random_2", "column": "random_2", "type": "category"},
        ],
        "output_features": [{"name": "Label", "column": "Label", "type": "binary"}],
        "trainer": {"epochs": 2, "batch_size": 8},
        "preprocessing": {},
    }
    df = pd.DataFrame(
        {
            "Index": np.arange(0, 200, 1),
            "random_1": np.random.randint(0, 50, 200),
            "random_2": np.random.choice(["Type A", "Type B", "Type C", "Type D"], 200),
            "Label": np.concatenate((np.zeros(180), np.ones(20))),
        }
    )

    config["preprocessing"][balance] = 0.5
    run_test_imbalance_local(df, config, balance)



from typing import Set

from ludwig.profiling.constants import CARDINALITY_EST, COUNTS_N, COUNTS_NULL, FREQUENT_ITEMS
from ludwig.profiling.types import ColumnProfileSummary
from ludwig.utils.audio_utils import is_audio_score
from ludwig.utils.image_utils import is_image_score


def get_num_nonnull_values(column_profile_summary: ColumnProfileSummary) -> int:
    """Returns the number of non-null values."""
    return column_profile_summary[COUNTS_N] - column_profile_summary[COUNTS_NULL]


def get_pct_null_values(column_profile_summary: ColumnProfileSummary) -> float:
    """Returns the percentage of null values."""
    return column_profile_summary[COUNTS_NULL] / column_profile_summary[COUNTS_N]


def get_num_distinct_values(column_profile_summary: ColumnProfileSummary) -> int:
    """Returns the number of distinct values."""
    return int(column_profile_summary[CARDINALITY_EST])


def get_distinct_values(column_profile_summary: ColumnProfileSummary) -> Set[str]:
    """Returns a list of distinct values."""
    if FREQUENT_ITEMS not in column_profile_summary:
        return {}
    frequent_items = column_profile_summary[FREQUENT_ITEMS]
    if not frequent_items:
        # Can be an empty list if the feature is non-string.
        return {}
    return {frequent_item.value for frequent_item in frequent_items}


def get_pct_distinct_values(column_profile_summary: ColumnProfileSummary) -> float:
    """Returns the percentage of distinct values."""
    return get_num_distinct_values(column_profile_summary) / column_profile_summary[COUNTS_N]


def get_distinct_values_balance(column_profile_summary: ColumnProfileSummary) -> float:
    """Returns the ratio of the least frequent / the most frequent item.

    The lower the value, the more imbalanced this feature is.
    """
    if FREQUENT_ITEMS not in column_profile_summary:
        return -1
    frequent_items = column_profile_summary[FREQUENT_ITEMS]
    if not frequent_items:
        # Can be an empty list if the feature is non-string.
        return -1

    max_occurence = frequent_items[0].est
    min_occurence = frequent_items[-1].est
    return min_occurence / max_occurence


def are_values_images(distinct_values: Set[str], feature_name: str) -> bool:
    """Returns whether the given values are probably images."""
    overall_image_score = 0
    for value in distinct_values:
        overall_image_score += is_image_score(None, value, column=feature_name)
        if overall_image_score > 3:
            return True

    if overall_image_score > 0.5 * len(distinct_values):
        return True
    return False


def are_values_audio(distinct_values: Set[str], feature_name: str) -> bool:
    """Returns whether the given values are probably audio."""
    overall_audio_score = 0
    for value in distinct_values:
        overall_audio_score += is_audio_score(value)
        if overall_audio_score > 3:
            return True

    if overall_audio_score > 0.5 * len(distinct_values):
        return True
    return False



#! /usr/bin/env python
# Copyright (c) 2022 Predibase, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

import logging
from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import List, Type, Union

import numpy as np
import torch
import torch.nn as nn

from ludwig.api_annotations import DeveloperAPI
from ludwig.constants import BINARY, CATEGORY
from ludwig.utils.registry import DEFAULT_KEYS, Registry

logger = logging.getLogger(__name__)

calibration_registry = Registry()


@DeveloperAPI
def register_calibration(name: str, features: Union[str, List[str]], default=False):
    """Registers a calibration implementation for a list of features."""
    if isinstance(features, str):
        features = [features]

    def wrap(cls):
        for feature in features:
            feature_registry = calibration_registry.get(feature, {})
            feature_registry[name] = cls
            if default:
                for key in DEFAULT_KEYS:
                    feature_registry[key] = cls
            calibration_registry[feature] = feature_registry
        return cls

    return wrap


@DeveloperAPI
def get_calibration_cls(feature: str, calibration_method: str) -> Type["CalibrationModule"]:
    """Get calibration class for specified feature type and calibration method."""
    if not calibration_method:
        return None
    if feature in calibration_registry:
        if calibration_method in calibration_registry[feature]:
            return calibration_registry[feature][calibration_method]
        else:
            raise ValueError(f"Calibration method {calibration_method} not supported for {feature} output features")
    else:
        raise ValueError(f"Calibration not yet supported for {feature} output features")
    return None


@DeveloperAPI
class ECELoss(nn.Module):
    """Calculates the Expected Calibration Error of a model.

    The input to this loss is the logits of a model, NOT the softmax scores.
    This divides the confidence outputs into equally-sized interval bins.
    In each bin, we compute the confidence gap:

        bin_gap = | avg_confidence_in_bin - accuracy_in_bin |

    We then return an average of the gaps, weighted by the number of samples in each bin.

    References:
        Naeini, Mahdi Pakdaman, Gregory F. Cooper, and Milos Hauskrecht
        "Obtaining Well Calibrated Probabilities Using Bayesian Binning." AAAI. 2015.

        Chuan Guo, Geoff Pleiss, Yu Sun, Kilian Q. Weinberger
        "On Calibration of Modern Neural Networks." PMLR 2017.
    """

    def __init__(self, n_bins: int = 15):
        """n_bins (int): number of confidence interval bins."""
        super().__init__()
        bin_boundaries = torch.linspace(0, 1, n_bins + 1)
        self.bin_lowers = bin_boundaries[:-1]
        self.bin_uppers = bin_boundaries[1:]

    def forward(self, logits: torch.Tensor, one_hot_labels: torch.Tensor) -> torch.Tensor:
        softmaxes = nn.functional.softmax(logits, dim=1)
        confidences, predictions = torch.max(softmaxes, 1)
        labels = torch.argmax(one_hot_labels, 1)
        accuracies = predictions.eq(labels)
        ece = torch.zeros(1, device=logits.device)
        for bin_lower, bin_upper in zip(self.bin_lowers, self.bin_uppers):
            # Calculates |confidence - accuracy| in each bin
            in_bin = confidences.gt(bin_lower.item()) * confidences.le(bin_upper.item())
            prop_in_bin = in_bin.float().mean()
            if prop_in_bin.item() > 0:
                accuracy_in_bin = accuracies[in_bin].float().mean()
                avg_confidence_in_bin = confidences[in_bin].mean()
                ece += torch.abs(avg_confidence_in_bin - accuracy_in_bin) * prop_in_bin
        return ece


@DeveloperAPI
@dataclass
class CalibrationResult:
    """Tracks results of probability calibration."""

    before_calibration_nll: float
    before_calibration_ece: float
    after_calibration_nll: float
    after_calibration_ece: float


@DeveloperAPI
class CalibrationModule(nn.Module, ABC):
    @abstractmethod
    def train_calibration(
        self, logits: Union[torch.Tensor, np.ndarray], labels: Union[torch.Tensor, np.ndarray]
    ) -> CalibrationResult:
        """Calibrate output probabilities using logits and labels from validation set."""
        return NotImplementedError()


@DeveloperAPI
@register_calibration("temperature_scaling", [BINARY, CATEGORY], default=True)
class TemperatureScaling(CalibrationModule):
    """Implements temperature scaling of logits. Based on results from "On Calibration of Modern Neural Networks":
    https://arxiv.org/abs/1706.04599. Temperature scaling scales all logits by the same constant factor. Though it
    may modify output probabilities it will never change argmax or categorical top-n predictions. In the case of
    binary classification with a threshold, however, calibration may change predictions.

    Implementation inspired by https://github.com/gpleiss/temperature_scaling

    Args:
        num_classes: The number of classes. Must be 2 if binary is True.
        binary: If binary is true, logits is expected to be a 1-dimensional array. If false, logits is a 2-dimensional
                array of shape (num_examples, num_classes).
    """

    def __init__(self, num_classes: int = 2, binary: bool = False):
        super().__init__()
        self.num_classes = 2 if binary else num_classes
        self.binary = binary
        self.device = "cuda" if torch.cuda.is_available() else "cpu"
        self.temperature = nn.Parameter(torch.ones(1), requires_grad=False).to(self.device)

    def train_calibration(
        self, logits: Union[torch.Tensor, np.ndarray], labels: Union[torch.Tensor, np.ndarray]
    ) -> CalibrationResult:
        logits = torch.as_tensor(logits, dtype=torch.float32, device=self.device)
        labels = torch.as_tensor(labels, dtype=torch.int64, device=self.device)
        one_hot_labels = nn.functional.one_hot(labels, self.num_classes).float()
        if self.binary:
            # Treat binary classification as multi-class with 2 classes to re-use code.
            # The math works out the same: softmax([0, a])[1] == sigmoid(a)
            logits = torch.stack([torch.zeros_like(logits), logits], axis=-1)
        nll_criterion = nn.CrossEntropyLoss().to(self.device)
        ece_criterion = ECELoss().to(self.device)
        # Saves the original temperature parameter, in case something goes wrong in optimization.
        original_temperature = self.temperature.clone().detach()
        self.temperature.requires_grad = True
        # Calculate NLL and ECE before temperature scaling
        before_calibration_nll = nll_criterion(logits, one_hot_labels).item()
        before_calibration_ece = ece_criterion(logits, one_hot_labels).item()
        logger.info(
            "Before temperature scaling:\n"
            "    Negative log-likelihood: %.3f\n"
            "    Expected Calibration Error: %.3f" % (before_calibration_nll, before_calibration_ece)
        )

        # Optimizes the temperature to minimize NLL
        optimizer = torch.optim.LBFGS([self.temperature], lr=0.01, max_iter=50, line_search_fn="strong_wolfe")

        def eval():
            optimizer.zero_grad()
            loss = nll_criterion(self.scale_logits(logits), one_hot_labels)
            loss.backward()
            return loss

        optimizer.step(eval)

        # Calculate NLL and ECE after temperature scaling
        after_calibration_nll = nll_criterion(self.scale_logits(logits), one_hot_labels).item()
        after_calibration_ece = ece_criterion(self.scale_logits(logits), one_hot_labels).item()
        logger.info("Optimal temperature: %.3f" % self.temperature.item())
        logger.info(
            "After temperature scaling:\n"
            "    Negative log-likelihood: %.3f\n"
            "    Expected Calibration Error: %.3f" % (after_calibration_nll, after_calibration_ece)
        )
        self.temperature.requires_grad = False
        # This should never happen, but if expected calibration error is higher after optimizing temperature, revert.
        if after_calibration_ece > before_calibration_ece:
            logger.warning(
                "Expected calibration error higher after scaling, "
                "reverting to temperature=%.3f." % original_temperature.item()
            )
            with torch.no_grad():
                self.temperature.data = original_temperature.data
        return CalibrationResult(
            before_calibration_nll, before_calibration_ece, after_calibration_nll, after_calibration_ece
        )

    def scale_logits(self, logits: torch.Tensor) -> torch.Tensor:
        return torch.div(logits, self.temperature)

    def forward(self, logits: torch.Tensor) -> torch.Tensor:
        """Converts logits to probabilities."""
        scaled_logits = self.scale_logits(logits)
        if self.binary:
            return torch.sigmoid(scaled_logits)
        else:
            return torch.softmax(scaled_logits, -1)


@DeveloperAPI
@register_calibration("matrix_scaling", CATEGORY, default=False)
class MatrixScaling(CalibrationModule):
    """Implements matrix scaling of logits, as described in Beyond temperature scaling: Obtaining well-calibrated
    multiclass probabilities with Dirichlet calibration https://arxiv.org/abs/1910.12656.

    Unlike temperature scaling which has only one free parameter, matrix scaling has n_classes x (n_classes + 1)
    parameters. Use this only with a large validation set, as matrix scaling has a tendency to overfit small datasets.
    Also, unlike temperature scaling, matrix scaling can change the argmax or top-n predictions.

    NOTE: Matrix Scaling is not exposed in the UI or config yet, though it may be in a future release after testing.

    Args:
    num_classes: The number of classes.
    off_diagonal_l2: The regularization weight for off-diagonal matrix entries.
    mu: The regularization weight for bias vector. Defaults to off_diagonal_l2 if not specified.
    """

    def __init__(self, num_classes: int = 2, off_diagonal_l2: float = 0.01, mu: float = None):
        super().__init__()
        self.num_classes = num_classes
        self.device = "cuda" if torch.cuda.is_available() else "cpu"
        self.w = nn.Parameter(torch.eye(self.num_classes), requires_grad=False).to(self.device)
        self.b = nn.Parameter(torch.zeros(self.num_classes), requires_grad=False).to(self.device)
        self.off_diagonal_l2 = off_diagonal_l2
        self.mu = off_diagonal_l2 if mu is None else mu

    def train_calibration(
        self, logits: Union[torch.Tensor, np.ndarray], labels: Union[torch.Tensor, np.ndarray]
    ) -> CalibrationResult:
        logits = torch.as_tensor(logits, dtype=torch.float32, device=self.device)
        labels = torch.as_tensor(labels, dtype=torch.int64, device=self.device)
        one_hot_labels = nn.functional.one_hot(labels, self.num_classes).float()
        nll_criterion = nn.CrossEntropyLoss().to(self.device)
        ece_criterion = ECELoss().to(self.device)
        self.w.requires_grad = True
        self.b.requires_grad = True
        # Calculate NLL and ECE before temperature scaling
        before_calibration_nll = nll_criterion(logits, one_hot_labels).item()
        before_calibration_ece = ece_criterion(logits, one_hot_labels).item()
        logger.info(
            "Before matrix scaling:\n"
            "    Negative log-likelihood: %.3f\n"
            "    Expected Calibration Error: %.3f" % (before_calibration_nll, before_calibration_ece)
        )

        # Optimizes the linear transform to minimize NLL
        optimizer = torch.optim.LBFGS([self.w, self.b], lr=0.001, max_iter=200, line_search_fn="strong_wolfe")

        def eval():
            optimizer.zero_grad()
            loss = nll_criterion(self.scale_logits(logits), one_hot_labels) + self.regularization_terms()
            loss.backward()
            return loss

        optimizer.step(eval)

        # Calculate NLL and ECE after matrix scaling
        after_calibration_nll = nll_criterion(self.scale_logits(logits), one_hot_labels).item()
        after_calibration_ece = ece_criterion(self.scale_logits(logits), one_hot_labels).item()
        logger.info(
            "After matrix scaling:\n"
            "    Negative log-likelihood: %.3f\n"
            "    Expected Calibration Error: %.3f" % (after_calibration_nll, after_calibration_ece)
        )
        self.w.requires_grad = False
        self.b.requires_grad = False
        # This should never happen, but if expected calibration error is higher after optimizing matrix, revert.
        if after_calibration_ece > before_calibration_ece:
            logger.warning("Expected calibration error higher after matrix scaling, reverting to identity.")
            with torch.no_grad():
                self.w.data = torch.eye(self.num_classes)
                self.b.data = torch.zeros(self.num_classes)
        return CalibrationResult(
            before_calibration_nll, before_calibration_ece, after_calibration_nll, after_calibration_ece
        )

    def regularization_terms(self) -> torch.Tensor:
        """Off-Diagonal and Intercept Regularisation (ODIR).

        Described in "Beyond temperature scaling: Obtaining well-calibrated multiclass probabilities with Dirichlet
        calibration" https://proceedings.neurips.cc/paper/2019/file/8ca01ea920679a0fe3728441494041b9-Paper.pdf
        """
        off_diagonal_entries = torch.masked_select(self.w, ~torch.eye(self.num_classes, dtype=bool))
        weight_matrix_loss = self.off_diagonal_l2 * torch.linalg.vector_norm(off_diagonal_entries)
        bias_vector_loss = self.mu * torch.linalg.vector_norm(self.b, 2)
        return bias_vector_loss + weight_matrix_loss

    def scale_logits(self, logits: torch.Tensor) -> torch.Tensor:
        return torch.matmul(self.w, logits.T).T + self.b

    def forward(self, logits: torch.Tensor) -> torch.Tensor:
        """Converts logits to probabilities."""
        return torch.softmax(self.scale_logits(logits), -1)



import pytest
import torch

from ludwig.modules.attention_modules import (
    FeedForwardAttentionReducer,
    MultiHeadSelfAttention,
    TransformerBlock,
    TransformerStack,
)
from ludwig.utils.misc_utils import set_random_seed
from tests.integration_tests.parameter_update_utils import check_module_parameters_updated

RANDOM_SEED = 1919


@pytest.mark.parametrize("input_hidden_size", [128, 256, 512])
@pytest.mark.parametrize("input_seq_size", [10, 20])
@pytest.mark.parametrize("input_batch_size", [16, 32])
def test_feed_forward_attention_reducer(input_batch_size: int, input_seq_size: int, input_hidden_size: int):
    # make repeatable
    set_random_seed(RANDOM_SEED)

    # Generate synthetic data
    current_inputs = torch.normal(0, 1, size=[input_batch_size, input_seq_size, input_hidden_size], dtype=torch.float32)

    # instantiate feed forward attention reducer
    feed_forward_attention_reducer = FeedForwardAttentionReducer(input_hidden_size)

    result = feed_forward_attention_reducer(current_inputs)

    # ensure returned tensor is the correct shape
    assert list(result.shape) == [input_batch_size, input_hidden_size]

    # check for parameter updating if fully connected layer is present
    target = torch.randn(result.shape)
    fpc, tpc, upc, not_updated = check_module_parameters_updated(
        feed_forward_attention_reducer,
        (current_inputs,),
        target,
    )
    assert upc == tpc, f"Some parameters not updated.  These parameters not updated: {not_updated}"


@pytest.mark.parametrize("input_hidden_size", [128, 256, 512])
@pytest.mark.parametrize("input_seq_size", [1, 10, 20])
@pytest.mark.parametrize("input_batch_size", [16, 32])
def test_multihead_self_attention(input_batch_size: int, input_seq_size: int, input_hidden_size: int):
    # make repeatable
    set_random_seed(RANDOM_SEED)

    # Generate synthetic data
    current_inputs = torch.normal(0, 1, size=[input_batch_size, input_seq_size, input_hidden_size], dtype=torch.float32)

    # instantiate feed forward attention reducer
    multihead_self_attention = MultiHeadSelfAttention(input_hidden_size, input_hidden_size)

    result = multihead_self_attention(current_inputs)

    # ensure returned tensor is the correct shape
    assert list(result.shape) == [input_batch_size, input_seq_size, input_hidden_size]

    # check for parameter updating if fully connected layer is present
    target = torch.randn(result.shape)
    fpc, tpc, upc, not_updated = check_module_parameters_updated(
        multihead_self_attention,
        (current_inputs,),
        target,
    )

    # adjustment required for single token sequence because self attention query and key parameters are
    # not updated
    single_sequence_token_adjustment = 4 if input_seq_size == 1 else 0
    assert upc == (
        tpc - single_sequence_token_adjustment
    ), f"Some parameters not updated.  These parameters not updated: {not_updated}"


# heads must be a divisor of input_hidden_size
@pytest.mark.parametrize("heads", [8, 16])
@pytest.mark.parametrize("output_size", [64, 128, 256])
@pytest.mark.parametrize("input_hidden_size", [128, 256, 512])
@pytest.mark.parametrize("input_seq_size", [10, 20])
@pytest.mark.parametrize("input_batch_size", [16, 32])
def test_transformer_block(
    input_batch_size: int,
    input_seq_size: int,
    input_hidden_size: int,
    output_size: int,
    heads: int,
):
    # make repeatable
    set_random_seed(RANDOM_SEED)

    # Generate synthetic data
    current_inputs = torch.normal(0, 1, size=[input_batch_size, input_seq_size, input_hidden_size], dtype=torch.float32)

    # instantiate feed forward attention reducer
    transformer_block = TransformerBlock(input_hidden_size, input_seq_size, input_hidden_size, heads, output_size)

    result = transformer_block(current_inputs)

    # ensure returned tensor is the correct shape
    assert list(result.shape) == [input_batch_size, input_seq_size, input_hidden_size]

    # check for parameter updating if fully connected layer is present
    target = torch.randn(result.shape)
    fpc, tpc, upc, not_updated = check_module_parameters_updated(
        transformer_block,
        (current_inputs,),
        target,
    )
    assert upc == tpc, f"Some parameters not updated.  These parameters not updated: {not_updated}"


@pytest.mark.parametrize("num_layers", [1, 4])
# heads must be a divisor of input_hidden_size
@pytest.mark.parametrize("heads", [8, 16])
@pytest.mark.parametrize("output_size", [64, 128, 256])
@pytest.mark.parametrize("input_hidden_size", [128, 256, 512])
@pytest.mark.parametrize("input_seq_size", [10, 20])
@pytest.mark.parametrize("input_batch_size", [16, 32])
def test_transformer_stack(
    input_batch_size: int,
    input_seq_size: int,
    input_hidden_size: int,
    output_size: int,
    heads: int,
    num_layers: int,
):
    # make repeatable
    set_random_seed(RANDOM_SEED)

    # Generate synthetic data
    current_inputs = torch.normal(0, 1, size=[input_batch_size, input_seq_size, input_hidden_size], dtype=torch.float32)

    # instantiate feed forward attention reducer
    transformer_stack = TransformerStack(
        input_hidden_size,
        input_seq_size,
        input_hidden_size,
        heads,
        output_size,
        num_layers,
    )

    result = transformer_stack(current_inputs)

    # ensure returned tensor is the correct shape
    assert list(result.shape) == [input_batch_size, input_seq_size, input_hidden_size]

    # check for parameter updating if fully connected layer is present
    target = torch.randn(result.shape)
    fpc, tpc, upc, not_updated = check_module_parameters_updated(
        transformer_stack,
        (current_inputs,),
        target,
    )
    assert upc == tpc, f"Some parameters not updated.  These parameters not updated: {not_updated}"



from typing import Dict

from ludwig.constants import TYPE
from ludwig.utils.registry import Registry

search_algorithm_registry = Registry()


def register_search_algorithm(name: str):
    def wrap(cls):
        search_algorithm_registry[name] = cls
        return cls

    return wrap


def get_search_algorithm_cls(name: str):
    return search_algorithm_registry[name]


def instantiate_search_algorithm(search_alg: Dict):
    search_alg_type = search_alg[TYPE]
    cls = get_search_algorithm_cls(search_alg_type)
    return cls(search_alg)



import uuid
from typing import Dict

import tqdm

try:
    from ray.air import session
except ImportError:
    session = None


class LudwigProgressBarActions:
    CREATE = "create"
    UPDATE = "update"
    CLOSE = "close"


class LudwigProgressBar:
    """Class for progress bars that supports distributed progress bars in ray.

    # Inputs

    :param report_to_ray: (bool) use the ray.air.session method
        to report progress to the ray driver. If false then this behaves as a normal tqdm
        progress bar
    :param config: (dict) the tqdm configs used for the progress bar. See https://github.com/tqdm/tqdm#parameters
        for list of parameters
    :param is_coordinator: (bool) whether the calling process is the coordinator process.

    # Example usage:

    ```python
    from ludwig.progress_bar import LudwigProgressBar

    config = {"total": 20, "desc": "Sample progress bar"}
    pbar = LudwigProgressBar(report_to_ray=False, config=config, is_coordinator=True)
    for i in range(20):
        pbar.update(1)
    pbar.close()
    ```
    """

    def __init__(
        self,
        report_to_ray: bool,
        config: Dict,
        is_coordinator: bool,
    ) -> None:
        """Constructor for the LudwigProgressBar class.

        # Inputs

        :param report_to_ray: (bool) use the ray.air.session method
            to report progress to the ray driver. If false then this behaves as a normal tqdm
            progress bar
        :param config: (dict) the tqdm configs used for the progress bar. See https://github.com/tqdm/tqdm#parameters
            for list of parameters
        :param is_coordinator: (bool) whether the calling process is the coordinator process.

        # Return

        :return: (None) `None`
        """
        if report_to_ray and session is None:
            raise ValueError("Set report_to_ray=True but ray is not installed. Run `pip install ray`")

        self.id = str(uuid.uuid4())[-8:]

        self.report_to_ray = report_to_ray
        self.is_coordinator = is_coordinator
        self.config = config

        self.total_steps = 0
        self.progress_bar = None

        if not self.report_to_ray:
            if self.is_coordinator:
                self.progress_bar = tqdm.tqdm(**config)
        else:
            if "file" in self.config:
                self.config.pop("file")
            # All processes need to call ray.train.report since ray has a lock that blocks
            # a process when calling report if there are processes that haven't called it. Similar
            # to a distributed checkpoint. Therefore we pass the flag to the driver
            session.report(
                metrics={
                    "progress_bar": {
                        "id": self.id,
                        "config": self.config,
                        "action": LudwigProgressBarActions.CREATE,
                        "is_coordinator": self.is_coordinator,
                    }
                }
            )

    def update(self, steps: int) -> None:
        """Updates the progress bar.

        # Inputs

        :param steps: (int) number of steps to update the progress bar by

        # Return

        :return: (None) `None`
        """
        self.total_steps += steps
        if self.progress_bar:
            self.progress_bar.update(steps)
        elif self.report_to_ray:
            session.report(
                metrics={
                    "progress_bar": {
                        "id": self.id,
                        "update_by": steps,
                        "is_coordinator": self.is_coordinator,
                        "action": LudwigProgressBarActions.UPDATE,
                    }
                }
            )

    def close(self) -> None:
        """Closes the progress bar.

        # Return

        :return: (None) `None`
        """
        if self.progress_bar:
            self.progress_bar.close()
        elif self.report_to_ray:
            session.report(
                metrics={
                    "progress_bar": {
                        "id": self.id,
                        "is_coordinator": self.is_coordinator,
                        "action": LudwigProgressBarActions.CLOSE,
                    }
                }
            )



#! /usr/bin/env python
# Copyright (c) 2019 Uber Technologies, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

import logging
from functools import partial
from typing import Dict, List, Union

import numpy as np
import torch

from ludwig.constants import (
    COLUMN,
    LAST_PREDICTIONS,
    LENGTHS,
    NAME,
    PREDICTIONS,
    PROBABILITIES,
    PROBABILITY,
    PROC_COLUMN,
    SEQUENCE,
)
from ludwig.features.base_feature import BaseFeatureMixin, InputFeature, OutputFeature, PredictModule
from ludwig.features.feature_utils import compute_sequence_probability, compute_token_probabilities
from ludwig.schema.features.sequence_feature import SequenceInputFeatureConfig, SequenceOutputFeatureConfig
from ludwig.types import (
    FeatureMetadataDict,
    FeaturePostProcessingOutputDict,
    PreprocessingConfigDict,
    TrainingSetMetadataDict,
)
from ludwig.utils import output_feature_utils
from ludwig.utils.math_utils import softmax
from ludwig.utils.strings_utils import (
    build_sequence_matrix,
    create_vocabulary,
    SpecialSymbol,
    START_SYMBOL,
    STOP_SYMBOL,
    UNKNOWN_SYMBOL,
)
from ludwig.utils.tokenizers import get_tokenizer_from_registry
from ludwig.utils.types import DataFrame, TorchscriptPreprocessingInput

logger = logging.getLogger(__name__)


class _SequencePreprocessing(torch.nn.Module):
    """Torchscript-enabled version of preprocessing done by SequenceFeatureMixin.add_feature_data."""

    def __init__(self, metadata: TrainingSetMetadataDict):
        super().__init__()
        self.lowercase = metadata["preprocessing"]["lowercase"]
        self.tokenizer_type = metadata["preprocessing"]["tokenizer"]
        self.tokenizer = get_tokenizer_from_registry(self.tokenizer_type)(
            pretrained_model_name_or_path=metadata["preprocessing"].get("pretrained_model_name_or_path", None)
        )

        if not isinstance(self.tokenizer, torch.nn.Module):
            raise ValueError(f"tokenizer must be a torch.nn.Module, got {self.tokenizer}")

        self.padding_symbol = metadata["preprocessing"]["padding_symbol"]
        self.unknown_symbol = metadata["preprocessing"]["unknown_symbol"]
        self.start_symbol = START_SYMBOL
        self.stop_symbol = STOP_SYMBOL
        self.max_sequence_length = int(metadata["max_sequence_length"])
        self.unit_to_id = metadata["str2idx"]
        self.computed_fill_value = metadata["preprocessing"]["computed_fill_value"]

    def forward(self, v: TorchscriptPreprocessingInput) -> torch.Tensor:
        """Takes a list of strings and returns a tensor of token ids."""
        if not torch.jit.isinstance(v, List[str]):
            raise ValueError(f"Unsupported input: {v}")

        futures: List[torch.jit.Future[torch.Tensor]] = []
        for sequence in v:
            futures.append(
                torch.jit.fork(
                    self._process_sequence,
                    sequence,
                )
            )

        sequence_matrix = []
        for future in futures:
            sequence_matrix.append(torch.jit.wait(future))

        return torch.stack(sequence_matrix)

    def _process_sequence(self, sequence: str) -> torch.Tensor:
        sequence = self.computed_fill_value if sequence == "nan" else sequence

        # If tokenizer is HF, we defer lowercase transformation to the tokenizer.
        if self.lowercase and self.tokenizer_type != "hf_tokenizer":
            sequence_str: str = sequence.lower()
        else:
            sequence_str: str = sequence

        sequence_vector = torch.full([self.max_sequence_length], self.unit_to_id[self.padding_symbol])

        if self.tokenizer_type == "hf_tokenizer":
            # Handles start, stop, and unknown symbols implicitly
            unit_sequence = self.tokenizer(sequence)
            assert torch.jit.isinstance(unit_sequence, List[int])
            # Ensures that the sequence lengths are aligned between the input and output tensors.
            sequence_length = min(len(unit_sequence), self.max_sequence_length)
            sequence_vector[:sequence_length] = torch.tensor(unit_sequence)[:sequence_length]
            return sequence_vector

        # If tokenizer is not HF, we manually convert tokens to IDs and insert start, stop, and unknown symbols.
        unit_sequence = self.tokenizer(sequence_str)
        assert torch.jit.isinstance(unit_sequence, List[str])

        sequence_vector[0] = self.unit_to_id[self.start_symbol]
        if len(unit_sequence) + 1 < self.max_sequence_length:
            sequence_length = len(unit_sequence)
            sequence_vector[len(unit_sequence) + 1] = self.unit_to_id[self.stop_symbol]
        else:
            sequence_length = self.max_sequence_length - 1

        for i in range(sequence_length):
            curr_unit = unit_sequence[i]
            if curr_unit in self.unit_to_id:
                curr_id = self.unit_to_id[curr_unit]
            else:
                curr_id = self.unit_to_id[self.unknown_symbol]
            sequence_vector[i + 1] = curr_id
        return sequence_vector


class _SequencePostprocessing(torch.nn.Module):
    def __init__(self, metadata: TrainingSetMetadataDict):
        super().__init__()
        self.max_sequence_length = int(metadata["max_sequence_length"])
        self.idx2str = metadata["idx2str"]
        self.unknown_symbol = UNKNOWN_SYMBOL
        self.predictions_key = PREDICTIONS
        self.probabilities_key = PROBABILITIES
        self.probability_key = PROBABILITY

    def forward(self, preds: Dict[str, torch.Tensor], feature_name: str) -> FeaturePostProcessingOutputDict:
        pred_predictions = output_feature_utils.get_output_feature_tensor(preds, feature_name, self.predictions_key)
        pred_probabilities = output_feature_utils.get_output_feature_tensor(preds, feature_name, self.probabilities_key)

        predictions: List[List[str]] = []
        for sequence in pred_predictions:
            sequence_predictions: List[str] = []
            for i in range(self.max_sequence_length):
                unit_id = int(sequence[i].item())
                if unit_id < len(self.idx2str):
                    unit_prediction = self.idx2str[unit_id]
                else:
                    unit_prediction = self.unknown_symbol
                sequence_predictions.append(unit_prediction)
            predictions.append(sequence_predictions)

        probabilities, _ = torch.max(pred_probabilities, dim=-1)
        probability = torch.sum(torch.log(probabilities), dim=-1)

        return {
            self.predictions_key: predictions,
            self.probabilities_key: probabilities,
            self.probability_key: probability,
        }


class _SequencePredict(PredictModule):
    def forward(self, inputs: Dict[str, torch.Tensor], feature_name: str) -> Dict[str, torch.Tensor]:
        logits = output_feature_utils.get_output_feature_tensor(inputs, feature_name, self.logits_key)
        probabilities = torch.softmax(logits, -1)
        predictions = torch.argmax(logits, -1)

        # predictions: [batch_size, sequence_length]
        # probabilities: [batch_size, sequence_length, vocab_size]
        # logits: [batch_size, sequence_length, vocab_size]
        return {self.predictions_key: predictions, self.probabilities_key: probabilities, self.logits_key: logits}


class SequenceFeatureMixin(BaseFeatureMixin):
    @staticmethod
    def type():
        return SEQUENCE

    @staticmethod
    def cast_column(column, backend):
        return column.astype(str)

    @staticmethod
    def get_feature_meta(
        column, preprocessing_parameters: PreprocessingConfigDict, backend, is_input_feature: bool
    ) -> FeatureMetadataDict:
        idx2str, str2idx, str2freq, max_length, _, _, _, _ = create_vocabulary(
            column,
            preprocessing_parameters["tokenizer"],
            lowercase=preprocessing_parameters["lowercase"],
            num_most_frequent=preprocessing_parameters["most_common"],
            vocab_file=preprocessing_parameters["vocab_file"],
            unknown_symbol=preprocessing_parameters["unknown_symbol"],
            padding_symbol=preprocessing_parameters["padding_symbol"],
            ngram_size=preprocessing_parameters["ngram_size"],
            processor=backend.df_engine,
        )
        logger.info(f"Max length of feature '{column.name}': {max_length} (without start and stop symbols)")

        # Use sequence_length if provided, otherwise use max length found in dataset.
        if preprocessing_parameters["sequence_length"] is not None:
            logger.info(
                f"Setting max length to sequence_length={preprocessing_parameters['sequence_length']} provided in "
                f"preprocessing parameters"
            )
            max_sequence_length = preprocessing_parameters["sequence_length"]
        else:
            max_sequence_length = max_length + 2  # For start and stop symbols.
            logger.info(f"Setting max length using dataset: {max_sequence_length} (including start and stop symbols)")

            # If max_sequence_length is None, then use the max length found in the dataset.
            if (
                preprocessing_parameters["max_sequence_length"] is not None
                and preprocessing_parameters["max_sequence_length"] < max_sequence_length
            ):
                logger.info(
                    f"Truncating max length with max_sequence_length={preprocessing_parameters['max_sequence_length']} "
                    f"from preprocessing parameters"
                )
                max_sequence_length = preprocessing_parameters["max_sequence_length"]

        logger.info(f"max sequence length is {max_sequence_length} for feature '{column.name}'")
        return {
            "idx2str": idx2str,
            "str2idx": str2idx,
            "str2freq": str2freq,
            "vocab_size": len(idx2str),
            "max_sequence_length": max_sequence_length,
        }

    @staticmethod
    def feature_data(column, metadata, preprocessing_parameters: PreprocessingConfigDict, backend):
        sequence_data = build_sequence_matrix(
            sequences=column,
            inverse_vocabulary=metadata["str2idx"],
            tokenizer_type=preprocessing_parameters["tokenizer"],
            length_limit=metadata["max_sequence_length"],
            padding_symbol=preprocessing_parameters["padding_symbol"],
            padding=preprocessing_parameters["padding"],
            unknown_symbol=preprocessing_parameters["unknown_symbol"],
            lowercase=preprocessing_parameters["lowercase"],
            tokenizer_vocab_file=preprocessing_parameters["vocab_file"],
            processor=backend.df_engine,
        )
        return sequence_data

    @staticmethod
    def add_feature_data(
        feature_config,
        input_df,
        proc_df,
        metadata,
        preprocessing_parameters: PreprocessingConfigDict,
        backend,
        skip_save_processed_input,
    ):
        sequence_data = SequenceInputFeature.feature_data(
            input_df[feature_config[COLUMN]],
            metadata[feature_config[NAME]],
            preprocessing_parameters,
            backend,
        )
        proc_df[feature_config[PROC_COLUMN]] = sequence_data
        return proc_df


class SequenceInputFeature(SequenceFeatureMixin, InputFeature):
    def __init__(self, input_feature_config: SequenceInputFeatureConfig, encoder_obj=None, **kwargs):
        super().__init__(input_feature_config, **kwargs)

        if encoder_obj:
            self.encoder_obj = encoder_obj
        else:
            self.encoder_obj = self.initialize_encoder(input_feature_config.encoder)

    def forward(self, inputs: torch.Tensor, mask=None):
        assert isinstance(inputs, torch.Tensor)
        assert inputs.dtype in [torch.int8, inputs.dtype, torch.int16, torch.int32, torch.int64]
        assert len(inputs.shape) == 2
        inputs_exp = inputs.type(torch.int32)
        inputs_mask = torch.not_equal(inputs, SpecialSymbol.PADDING.value)
        lengths = torch.sum(inputs_mask.type(torch.int32), dim=1)
        encoder_output = self.encoder_obj(inputs_exp, mask=inputs_mask)
        encoder_output[LENGTHS] = lengths
        return encoder_output

    @property
    def input_dtype(self):
        return torch.int32

    @staticmethod
    def update_config_with_metadata(feature_config, feature_metadata, *args, **kwargs):
        feature_config.encoder.vocab = feature_metadata["idx2str"]
        feature_config.encoder.vocab_size = len(feature_metadata["idx2str"])
        feature_config.encoder.max_sequence_length = feature_metadata["max_sequence_length"]

    @staticmethod
    def get_schema_cls():
        return SequenceInputFeatureConfig

    @property
    def input_shape(self) -> torch.Size:
        return torch.Size([self.encoder_obj.config.max_sequence_length])

    @property
    def output_shape(self) -> torch.Size:
        return self.encoder_obj.output_shape

    @staticmethod
    def create_preproc_module(metadata: TrainingSetMetadataDict) -> torch.nn.Module:
        return _SequencePreprocessing(metadata)


class SequenceOutputFeature(SequenceFeatureMixin, OutputFeature):
    def __init__(
        self,
        output_feature_config: Union[SequenceOutputFeatureConfig, Dict],
        output_features: Dict[str, OutputFeature],
        **kwargs,
    ):
        super().__init__(output_feature_config, output_features, **kwargs)
        self.decoder_obj = self.initialize_decoder(output_feature_config.decoder)
        self._setup_loss()
        self._setup_metrics()

    def logits(self, inputs: Dict[str, torch.Tensor], target=None):
        return self.decoder_obj(inputs, target=target)

    def create_predict_module(self) -> PredictModule:
        return _SequencePredict()

    def get_prediction_set(self):
        return self.decoder_obj.get_prediction_set()

    @classmethod
    def get_output_dtype(cls):
        return torch.int32

    @property
    def input_shape(self) -> torch.Size:
        # Dummy implementation.
        return torch.Size([1])

    @property
    def output_shape(self) -> torch.Size:
        return torch.Size([self.decoder_obj.config.max_sequence_length])

    @staticmethod
    def update_config_with_metadata(feature_config, feature_metadata, *args, **kwargs):
        feature_config.decoder.vocab_size = feature_metadata["vocab_size"]
        feature_config.decoder.max_sequence_length = feature_metadata["max_sequence_length"]
        if isinstance(feature_config.loss.class_weights, (list, tuple)):
            if len(feature_config.loss.class_weights) != feature_config.decoder.vocab_size:
                raise ValueError(
                    "The length of class_weights ({}) is not compatible with "
                    "the number of classes ({}) for feature {}. "
                    "Check the metadata JSON file to see the classes "
                    "and their order and consider there needs to be a weight "
                    "for the <UNK> and <PAD> class too.".format(
                        len(feature_config.loss.class_weights),
                        feature_config.decoder.vocab_size,
                        feature_config.column,
                    )
                )

        if feature_config.loss.class_similarities_temperature > 0:
            if "class_similarities" in feature_config.loss:
                similarities = feature_config.loss.class_similarities
                temperature = feature_config.loss.class_similarities_temperature

                curr_row = 0
                first_row_length = 0
                is_first_row = True
                for row in similarities:
                    if is_first_row:
                        first_row_length = len(row)
                        is_first_row = False
                        curr_row += 1
                    else:
                        curr_row_length = len(row)
                        if curr_row_length != first_row_length:
                            raise ValueError(
                                "The length of row {} of the class_similarities "
                                "of {} is {}, different from the length of "
                                "the first row {}. All rows must have "
                                "the same length.".format(
                                    curr_row, feature_config.column, curr_row_length, first_row_length
                                )
                            )
                        else:
                            curr_row += 1
                all_rows_length = first_row_length

                if all_rows_length != len(similarities):
                    raise ValueError(
                        "The class_similarities matrix of {} has "
                        "{} rows and {} columns, "
                        "their number must be identical.".format(
                            feature_config.column, len(similarities), all_rows_length
                        )
                    )

                if all_rows_length != feature_config.decoder.vocab_size:
                    raise ValueError(
                        "The size of the class_similarities matrix of {} is "
                        "{}, different from the number of classes ({}). "
                        "Check the metadata JSON file to see the classes "
                        "and their order and "
                        "consider <UNK> and <PAD> class too.".format(
                            feature_config.column, all_rows_length, feature_config.decoder.vocab_size
                        )
                    )

                similarities = np.array(similarities, dtype=np.float32)
                for i in range(len(similarities)):
                    similarities[i, :] = softmax(similarities[i, :], temperature=temperature)
                feature_config.loss.class_similarities = similarities
            else:
                raise ValueError(
                    "class_similarities_temperature > 0, "
                    "but no class_similarities are provided "
                    "for feature {}".format(feature_config.column)
                )

    @staticmethod
    def calculate_overall_stats(predictions, targets, train_set_metadata):
        # TODO(Justin): Add a confusion matrix, see
        # https://github.com/ludwig-ai/ludwig/blob/tf-legacy/ludwig/features/sequence_feature.py#L411
        return {}

    def postprocess_predictions(
        self,
        result,
        metadata,
    ):
        predictions_col = f"{self.feature_name}_{PREDICTIONS}"
        lengths_col = f"{self.feature_name}_{LENGTHS}"
        if predictions_col in result:
            if "idx2str" in metadata:

                def idx2str(row):
                    pred = row[predictions_col]
                    length = metadata["max_sequence_length"]
                    return [
                        metadata["idx2str"][token] if token < len(metadata["idx2str"]) else UNKNOWN_SYMBOL
                        for token in [pred[i] for i in range(length)]
                    ]

                result[predictions_col] = result.apply(idx2str, axis=1)

        last_preds_col = f"{self.feature_name}_{LAST_PREDICTIONS}"
        if last_preds_col in result:
            if "idx2str" in metadata:

                def last_idx2str(last_pred):
                    if last_pred < len(metadata["idx2str"]):
                        return metadata["idx2str"][last_pred]
                    return UNKNOWN_SYMBOL

                result[last_preds_col] = result[last_preds_col].map(last_idx2str)

        probs_col = f"{self.feature_name}_{PROBABILITIES}"
        prob_col = f"{self.feature_name}_{PROBABILITY}"
        if probs_col in result:
            # currently does not return full probabilties because usually it is huge:
            # dataset x length x classes
            # TODO: add a mechanism for letting the user decide to save it
            result[probs_col] = result[probs_col].map(compute_token_probabilities)
            result[prob_col] = result[probs_col].map(
                partial(
                    compute_sequence_probability,
                    max_sequence_length=metadata["max_sequence_length"],
                    return_log_prob=True,
                )
            )

        if lengths_col in result:
            del result[lengths_col]

        return result

    @staticmethod
    def create_postproc_module(metadata: TrainingSetMetadataDict) -> torch.nn.Module:
        return _SequencePostprocessing(metadata)

    @staticmethod
    def get_schema_cls():
        return SequenceOutputFeatureConfig

    def flatten(self, df: DataFrame) -> DataFrame:
        probs_col = f"{self.feature_name}_{PROBABILITIES}"
        df[probs_col] = df[probs_col].apply(lambda x: x.flatten())
        return df

    def unflatten(self, df: DataFrame) -> DataFrame:
        probs_col = f"{self.feature_name}_{PROBABILITIES}"
        df[probs_col] = df[probs_col].apply(
            lambda x: x.reshape(-1, self.decoder_obj.config.num_classes), meta=(probs_col, "object")
        )
        return df



from typing import Dict, List, Type, Union

from ludwig.api_annotations import DeveloperAPI
from ludwig.constants import (
    BINARY,
    BINARY_WEIGHTED_CROSS_ENTROPY,
    CATEGORY,
    MEAN_ABSOLUTE_ERROR,
    MEAN_SQUARED_ERROR,
    NUMBER,
    ROOT_MEAN_SQUARED_ERROR,
    ROOT_MEAN_SQUARED_PERCENTAGE_ERROR,
    SEQUENCE,
    SEQUENCE_SOFTMAX_CROSS_ENTROPY,
    SET,
    SIGMOID_CROSS_ENTROPY,
    SOFTMAX_CROSS_ENTROPY,
    TEXT,
    TIMESERIES,
    VECTOR,
)
from ludwig.schema import utils as schema_utils
from ludwig.schema.metadata import LOSS_METADATA
from ludwig.schema.utils import ludwig_dataclass
from ludwig.utils.registry import Registry

ROBUST_LAMBDA_DESCRIPTION = (
    "Replaces the loss with `(1 - robust_lambda) * loss + robust_lambda / c` where `c` is the number of "
    "classes. Useful in case of noisy labels."
)

CONFIDENCE_PENALTY_DESCRIPTION = (
    "Penalizes overconfident predictions (low entropy) by adding an additional term "
    "that penalizes too confident predictions by adding a `a * (max_entropy - entropy) / max_entropy` "
    "term to the loss, where a is the value of this parameter. Useful in case of noisy labels."
)

CLASS_WEIGHTS_DESCRIPTION = (
    "Weights to apply to each class in the loss. If not specified, all classes are weighted equally. "
    "The value can be a vector of weights, one for each class, that is multiplied to the "
    "loss of the datapoints that have that class as ground truth. It is an alternative to oversampling in "
    "case of unbalanced class distribution. The ordering of the vector follows the category to integer ID "
    "mapping in the JSON metadata file (the `<UNK>` class needs to be included too). Alternatively, the value "
    "can be a dictionary with class strings as keys and weights as values, like "
    "`{class_a: 0.5, class_b: 0.7, ...}`."
)

CLASS_SIMILARITIES_DESCRIPTION = (
    "If not `null` it is a `c x c` matrix in the form of a list of lists that contains the mutual similarity of "
    "classes. It is used if `class_similarities_temperature` is greater than 0. The ordering of the vector follows "
    "the category to integer ID mapping in the JSON metadata file (the `<UNK>` class needs to be included too)."
)

CLASS_SIMILARITIES_TEMPERATURE_DESCRIPTION = (
    "The temperature parameter of the softmax that is performed on each row of `class_similarities`. The output of "
    "that softmax is used to determine the supervision vector to provide instead of the one hot vector that would be "
    "provided otherwise for each datapoint. The intuition behind it is that errors between similar classes are more "
    "tolerable than errors between really different classes."
)


@DeveloperAPI
@ludwig_dataclass
class BaseLossConfig(schema_utils.BaseMarshmallowConfig):
    """Base class for feature configs."""

    type: str

    weight: float = 1.0

    @classmethod
    def name(cls) -> str:
        return "[undefined]"


_loss_registry = Registry[Type[BaseLossConfig]]()
_loss_feature_registry = Registry[Dict[str, Type[BaseLossConfig]]]()


@DeveloperAPI
def get_loss_schema_registry() -> Registry[Type[BaseLossConfig]]:
    return _loss_registry


@DeveloperAPI
def get_loss_cls(feature: str, name: str) -> Type[BaseLossConfig]:
    return _loss_feature_registry[feature][name]


@DeveloperAPI
def get_loss_classes(feature: str) -> Dict[str, Type[BaseLossConfig]]:
    return _loss_feature_registry[feature]


def register_loss(features: Union[str, List[str]]):
    if isinstance(features, str):
        features = [features]

    def wrap(cls: Type[BaseLossConfig]):
        _loss_registry[cls.type] = cls
        for feature in features:
            feature_registry = _loss_feature_registry.get(feature, {})
            feature_registry[cls.type] = cls
            _loss_feature_registry[feature] = feature_registry
        return cls

    return wrap


@DeveloperAPI
@register_loss([NUMBER, TIMESERIES, VECTOR])
@ludwig_dataclass
class MSELossConfig(BaseLossConfig):
    type: str = schema_utils.ProtectedString(
        MEAN_SQUARED_ERROR,
        description="Type of loss.",
    )

    weight: float = schema_utils.NonNegativeFloat(
        default=1.0,
        description="Weight of the loss.",
        parameter_metadata=LOSS_METADATA["MSELoss"]["weight"],
    )

    @classmethod
    def name(self) -> str:
        return "Mean Squared Error (MSE)"


@DeveloperAPI
@register_loss([NUMBER, TIMESERIES, VECTOR])
@ludwig_dataclass
class MAELossConfig(BaseLossConfig):
    type: str = schema_utils.ProtectedString(
        MEAN_ABSOLUTE_ERROR,
        description="Type of loss.",
    )

    weight: float = schema_utils.NonNegativeFloat(
        default=1.0,
        description="Weight of the loss.",
        parameter_metadata=LOSS_METADATA["MAELoss"]["weight"],
    )

    @classmethod
    def name(self) -> str:
        return "Mean Absolute Error (MAE)"


@DeveloperAPI
@register_loss([NUMBER])
@ludwig_dataclass
class RMSELossConfig(BaseLossConfig):
    type: str = schema_utils.ProtectedString(
        ROOT_MEAN_SQUARED_ERROR,
        description="Type of loss.",
    )

    weight: float = schema_utils.NonNegativeFloat(
        default=1.0,
        description="Weight of the loss.",
        parameter_metadata=LOSS_METADATA["RMSELoss"]["weight"],
    )

    @classmethod
    def name(self) -> str:
        return "Root Mean Squared Error (RMSE)"


@DeveloperAPI
@register_loss([NUMBER])
@ludwig_dataclass
class RMSPELossConfig(BaseLossConfig):
    type: str = schema_utils.ProtectedString(
        ROOT_MEAN_SQUARED_PERCENTAGE_ERROR,
        description="Type of loss.",
    )

    weight: float = schema_utils.NonNegativeFloat(
        default=1.0,
        description="Weight of the loss.",
        parameter_metadata=LOSS_METADATA["RMSPELoss"]["weight"],
    )

    @classmethod
    def name(self) -> str:
        return "Root Mean Squared Percentage Error (RMSPE)"


@DeveloperAPI
@register_loss([BINARY])
@ludwig_dataclass
class BWCEWLossConfig(BaseLossConfig):
    type: str = schema_utils.ProtectedString(
        BINARY_WEIGHTED_CROSS_ENTROPY,
        description="Type of loss.",
    )

    positive_class_weight: float = schema_utils.NonNegativeFloat(
        default=None,
        allow_none=True,
        description="Weight of the positive class.",
        parameter_metadata=LOSS_METADATA["BWCEWLoss"]["positive_class_weight"],
    )

    robust_lambda: int = schema_utils.NonNegativeInteger(
        default=0,
        description=ROBUST_LAMBDA_DESCRIPTION,
        parameter_metadata=LOSS_METADATA["BWCEWLoss"]["robust_lambda"],
    )

    confidence_penalty: float = schema_utils.NonNegativeFloat(
        default=0,
        description=CONFIDENCE_PENALTY_DESCRIPTION,
        parameter_metadata=LOSS_METADATA["BWCEWLoss"]["confidence_penalty"],
    )

    weight: float = schema_utils.NonNegativeFloat(
        default=1.0,
        description="Weight of the loss.",
        parameter_metadata=LOSS_METADATA["BWCEWLoss"]["weight"],
    )

    @classmethod
    def name(self) -> str:
        return "Binary Weighted Cross Entropy (BWCE)"


@DeveloperAPI
@register_loss([CATEGORY, VECTOR])
@ludwig_dataclass
class SoftmaxCrossEntropyLossConfig(BaseLossConfig):
    type: str = schema_utils.ProtectedString(
        SOFTMAX_CROSS_ENTROPY,
        description="Type of loss.",
    )

    class_weights: Union[List[float], float, None] = schema_utils.List(
        list_type=float,
        default=None,
        description=CLASS_WEIGHTS_DESCRIPTION,
        parameter_metadata=LOSS_METADATA["SoftmaxCrossEntropyLoss"]["class_weights"],
    )

    robust_lambda: int = schema_utils.NonNegativeInteger(
        default=0,
        description=ROBUST_LAMBDA_DESCRIPTION,
        parameter_metadata=LOSS_METADATA["SoftmaxCrossEntropyLoss"]["robust_lambda"],
    )

    confidence_penalty: float = schema_utils.NonNegativeFloat(
        default=0,
        description=CONFIDENCE_PENALTY_DESCRIPTION,
        parameter_metadata=LOSS_METADATA["SoftmaxCrossEntropyLoss"]["confidence_penalty"],
    )

    class_similarities: list = schema_utils.List(
        list,
        default=None,
        description=CLASS_SIMILARITIES_DESCRIPTION,
        parameter_metadata=LOSS_METADATA["SoftmaxCrossEntropyLoss"]["class_similarities"],
    )

    class_similarities_temperature: int = schema_utils.NonNegativeInteger(
        default=0,
        description=CLASS_SIMILARITIES_TEMPERATURE_DESCRIPTION,
        parameter_metadata=LOSS_METADATA["SoftmaxCrossEntropyLoss"]["class_similarities_temperature"],
    )

    weight: float = schema_utils.NonNegativeFloat(
        default=1.0,
        description="Weight of the loss.",
        parameter_metadata=LOSS_METADATA["SoftmaxCrossEntropyLoss"]["weight"],
    )

    @classmethod
    def name(self) -> str:
        return "Softmax Cross Entropy"


@DeveloperAPI
@register_loss([SEQUENCE, TEXT])
@ludwig_dataclass
class SequenceSoftmaxCrossEntropyLossConfig(BaseLossConfig):
    type: str = schema_utils.ProtectedString(
        SEQUENCE_SOFTMAX_CROSS_ENTROPY,
        description="Type of loss.",
    )

    class_weights: Union[List[float], float, None] = schema_utils.List(
        list_type=float,
        default=None,
        description=CLASS_WEIGHTS_DESCRIPTION,
        parameter_metadata=LOSS_METADATA["SequenceSoftmaxCrossEntropyLoss"]["class_weights"],
    )

    robust_lambda: int = schema_utils.NonNegativeInteger(
        default=0,
        description=ROBUST_LAMBDA_DESCRIPTION,
        parameter_metadata=LOSS_METADATA["SequenceSoftmaxCrossEntropyLoss"]["robust_lambda"],
    )

    confidence_penalty: float = schema_utils.NonNegativeFloat(
        default=0,
        description=CONFIDENCE_PENALTY_DESCRIPTION,
        parameter_metadata=LOSS_METADATA["SequenceSoftmaxCrossEntropyLoss"]["confidence_penalty"],
    )

    class_similarities: list = schema_utils.List(
        list,
        default=None,
        description=CLASS_SIMILARITIES_DESCRIPTION,
        parameter_metadata=LOSS_METADATA["SequenceSoftmaxCrossEntropyLoss"]["class_similarities"],
    )

    class_similarities_temperature: int = schema_utils.NonNegativeInteger(
        default=0,
        description=CLASS_SIMILARITIES_TEMPERATURE_DESCRIPTION,
        parameter_metadata=LOSS_METADATA["SequenceSoftmaxCrossEntropyLoss"]["class_similarities_temperature"],
    )

    weight: float = schema_utils.NonNegativeFloat(
        default=1.0,
        description="Weight of the loss.",
        parameter_metadata=LOSS_METADATA["SequenceSoftmaxCrossEntropyLoss"]["weight"],
    )

    unique: bool = schema_utils.Boolean(
        default=False,
        description="If true, the loss is only computed for unique elements in the sequence.",
        parameter_metadata=LOSS_METADATA["SequenceSoftmaxCrossEntropyLoss"]["unique"],
    )

    @classmethod
    def name(self) -> str:
        return "Sequence Softmax Cross Entropy"


@DeveloperAPI
@register_loss([SET])
@ludwig_dataclass
class SigmoidCrossEntropyLossConfig(BaseLossConfig):
    type: str = schema_utils.ProtectedString(
        SIGMOID_CROSS_ENTROPY,
        description="Type of loss.",
    )

    class_weights: Union[List[float], float, None] = schema_utils.List(
        list_type=float,
        default=None,
        description=CLASS_WEIGHTS_DESCRIPTION,
        parameter_metadata=LOSS_METADATA["SigmoidCrossEntropyLoss"]["class_weights"],
    )

    weight: float = schema_utils.NonNegativeFloat(
        default=1.0,
        description="Weight of the loss.",
        parameter_metadata=LOSS_METADATA["SigmoidCrossEntropyLoss"]["weight"],
    )

    @classmethod
    def name(self) -> str:
        return "Sigmoid Cross Entropy"



import json
import logging
import os.path
import re

import numpy as np
import pandas as pd
import pytest
import torch

from ludwig import globals as global_vars
from ludwig.api import LudwigModel
from ludwig.backend import LOCAL_BACKEND
from ludwig.constants import (
    BATCH_SIZE,
    CATEGORY,
    DEFAULTS,
    EPOCHS,
    INPUT_FEATURES,
    OUTPUT_FEATURES,
    PREPROCESSING,
    TRAINER,
    TRAINING,
)
from ludwig.contribs.mlflow import MlflowCallback
from ludwig.experiment import experiment_cli
from ludwig.features.number_feature import numeric_transformation_registry
from ludwig.globals import DESCRIPTION_FILE_NAME, TRAINING_PREPROC_FILE_NAME
from ludwig.schema.optimizers import optimizer_registry
from ludwig.utils.data_utils import load_json, replace_file_extension
from ludwig.utils.misc_utils import get_from_registry
from ludwig.utils.package_utils import LazyLoader
from tests.integration_tests import synthetic_test_data
from tests.integration_tests.utils import category_feature, generate_data, LocalTestBackend

mlflow = LazyLoader("mlflow", globals(), "mlflow")

RANDOM_SEED = 42


@pytest.mark.parametrize("early_stop", [3, 5])
def test_early_stopping(early_stop, tmp_path):
    input_features, output_features = synthetic_test_data.get_feature_configs()

    config = {
        "input_features": input_features,
        "output_features": output_features,
        "combiner": {"type": "concat"},
        TRAINER: {"epochs": 75, "early_stop": early_stop, "batch_size": 16},
    }

    # create sub-directory to store results
    results_dir = tmp_path / "results"
    results_dir.mkdir()

    # run experiment
    generated_data = synthetic_test_data.get_generated_data()
    _, _, _, _, output_dir = experiment_cli(
        training_set=generated_data.train_df,
        validation_set=generated_data.validation_df,
        test_set=generated_data.test_df,
        output_directory=str(results_dir),
        config=config,
        skip_save_processed_input=True,
        skip_save_progress=True,
        skip_save_unprocessed_output=True,
        skip_save_model=True,
        skip_save_log=True,
    )

    # test existence of required files
    train_stats_fp = os.path.join(output_dir, "training_statistics.json")
    metadata_fp = os.path.join(output_dir, DESCRIPTION_FILE_NAME)
    assert os.path.isfile(train_stats_fp)
    assert os.path.isfile(metadata_fp)

    # retrieve results so we can validate early stopping
    with open(train_stats_fp) as f:
        train_stats = json.load(f)
    with open(metadata_fp) as f:
        metadata = json.load(f)

    # get early stopping value
    early_stop_value = metadata["config"][TRAINER]["early_stop"]

    # retrieve validation losses
    vald_losses_data = train_stats["validation"]["combined"]["loss"]

    last_evaluation = len(vald_losses_data) - 1
    best_evaluation = np.argmin(vald_losses_data)

    assert last_evaluation - best_evaluation == early_stop_value


@pytest.mark.parametrize("skip_save_progress", [False])
@pytest.mark.parametrize("skip_save_model", [False, True])
def test_model_progress_save(skip_save_progress, skip_save_model, tmp_path):
    input_features, output_features = synthetic_test_data.get_feature_configs()

    config = {
        "input_features": input_features,
        "output_features": output_features,
        "combiner": {"type": "concat"},
        TRAINER: {"epochs": 5, BATCH_SIZE: 128},
    }

    # create sub-directory to store results
    results_dir = tmp_path / "results"
    results_dir.mkdir()

    # run experiment
    generated_data = synthetic_test_data.get_generated_data()
    _, _, _, _, output_dir = experiment_cli(
        training_set=generated_data.train_df,
        validation_set=generated_data.validation_df,
        test_set=generated_data.test_df,
        output_directory=str(results_dir),
        config=config,
        skip_save_processed_input=True,
        skip_save_progress=skip_save_progress,
        skip_save_unprocessed_output=True,
        skip_save_model=skip_save_model,
        skip_save_log=True,
    )

    # ========== Check for required result data sets =============
    model_dir = os.path.join(output_dir, "model")
    files = [f for f in os.listdir(model_dir) if re.match(r"model_weights", f)]
    if skip_save_model:
        assert len(files) == 0
    else:
        assert len(files) == 1

    training_checkpoints_dir = os.path.join(output_dir, "model", "training_checkpoints")
    training_checkpoints = os.listdir(training_checkpoints_dir)
    if skip_save_progress:
        assert len(training_checkpoints) == 0
    else:
        assert len(training_checkpoints) > 0


@pytest.mark.parametrize("optimizer", ["sgd", "adam"])
def test_resume_training(optimizer, tmp_path):
    input_features, output_features = synthetic_test_data.get_feature_configs()
    config = {
        "input_features": input_features,
        "output_features": output_features,
        "combiner": {"type": "concat"},
        TRAINER: {"epochs": 2, "batch_size": 16, "optimizer": {"type": optimizer}},
    }

    # create sub-directory to store results
    results_dir = tmp_path / "results"
    results_dir.mkdir()

    generated_data = synthetic_test_data.get_generated_data()
    _, _, _, _, output_dir1 = experiment_cli(
        config,
        training_set=generated_data.train_df,
        validation_set=generated_data.validation_df,
        test_set=generated_data.test_df,
    )

    config[TRAINER]["epochs"] = 5

    experiment_cli(
        config,
        training_set=generated_data.train_df,
        validation_set=generated_data.validation_df,
        test_set=generated_data.test_df,
        model_resume_path=output_dir1,
    )

    _, _, _, _, output_dir2 = experiment_cli(
        config,
        training_set=generated_data.train_df,
        validation_set=generated_data.validation_df,
        test_set=generated_data.test_df,
    )

    # compare learning curves with and without resuming
    ts1 = load_json(os.path.join(output_dir1, "training_statistics.json"))
    ts2 = load_json(os.path.join(output_dir2, "training_statistics.json"))
    print("ts1", ts1)
    print("ts2", ts2)
    assert ts1[TRAINING]["combined"]["loss"] == ts2[TRAINING]["combined"]["loss"]

    # compare predictions with and without resuming
    y_pred1 = np.load(os.path.join(output_dir1, "y_predictions.npy"))
    y_pred2 = np.load(os.path.join(output_dir2, "y_predictions.npy"))
    print("y_pred1", y_pred1)
    print("y_pred2", y_pred2)
    assert np.all(np.isclose(y_pred1, y_pred2))


@pytest.mark.parametrize("optimizer", ["sgd", "adam"])
def test_resume_training_mlflow(optimizer, tmp_path):
    input_features, output_features = synthetic_test_data.get_feature_configs()
    config = {
        "input_features": input_features,
        "output_features": output_features,
        "combiner": {"type": "concat"},
        TRAINER: {"epochs": 2, "batch_size": 16, "optimizer": {"type": optimizer}},
    }

    # create sub-directory to store results
    results_dir = tmp_path / "results"
    results_dir.mkdir()
    mlflow_uri = f"file://{tmp_path}/mlruns"
    experiment_name = optimizer + "_experiment"

    generated_data = synthetic_test_data.get_generated_data()
    _, _, _, _, output_dir1 = experiment_cli(
        config,
        training_set=generated_data.train_df,
        validation_set=generated_data.validation_df,
        test_set=generated_data.test_df,
        callbacks=[MlflowCallback(mlflow_uri)],
        experiment_name=experiment_name,
    )
    # Can't change any artifact spec on a run once it has been logged to mlflow, so skipping changing epochs

    _, _, _, _, output_dir2 = experiment_cli(
        config,
        training_set=generated_data.train_df,
        validation_set=generated_data.validation_df,
        test_set=generated_data.test_df,
        model_resume_path=output_dir1,
        callbacks=[MlflowCallback(mlflow_uri)],
        experiment_name=experiment_name,
    )

    # make sure there is only one mlflow run id
    experiment = mlflow.get_experiment_by_name(experiment_name)
    previous_runs = mlflow.search_runs([experiment.experiment_id])
    assert len(previous_runs) == 1


@pytest.mark.parametrize("optimizer_type", optimizer_registry)
def test_optimizers(optimizer_type, tmp_path):
    input_features, output_features = synthetic_test_data.get_feature_configs()

    config = {
        "input_features": input_features,
        "output_features": output_features,
        "combiner": {"type": "concat"},
        TRAINER: {"epochs": 5, "batch_size": 16, "evaluate_training_set": True, "optimizer": {"type": optimizer_type}},
    }

    # special handling for adadelta and lbfgs, break out of local minima
    if optimizer_type == "adadelta":
        config[TRAINER]["learning_rate"] = 0.1
    if optimizer_type == "lbfgs":
        config[TRAINER]["learning_rate"] = 0.05

    model = LudwigModel(config)

    # create sub-directory to store results
    results_dir = tmp_path / "results"
    results_dir.mkdir()

    # run experiment
    generated_data = synthetic_test_data.get_generated_data_for_optimizer()
    train_stats, preprocessed_data, output_directory = model.train(
        training_set=generated_data.train_df,
        output_directory=str(results_dir),
        config=config,
        skip_save_processed_input=True,
        skip_save_progress=True,
        skip_save_unprocessed_output=True,
        skip_save_model=True,
        skip_save_log=True,
    )

    # retrieve training losses for first and last entries.
    train_losses = train_stats[TRAINING]["combined"]["loss"]
    last_entry = len(train_losses)

    # ensure train loss for last entry is less than to the first entry.
    np.testing.assert_array_less(train_losses[last_entry - 1], train_losses[0])


def test_regularization(tmp_path):
    input_features, output_features = synthetic_test_data.get_feature_configs()

    config = {
        "input_features": input_features,
        "output_features": output_features,
        "combiner": {"type": "concat"},
        TRAINER: {
            "epochs": 1,
            "batch_size": 16,
            "regularization_lambda": 1,
        },
    }

    # create sub-directory to store results
    results_dir = tmp_path / "results"
    results_dir.mkdir()

    regularization_losses = []
    generated_data = synthetic_test_data.get_generated_data()
    for regularizer in [None, "l1", "l2", "l1_l2"]:
        np.random.seed(RANDOM_SEED)
        torch.manual_seed(RANDOM_SEED)

        # setup regularization parameters
        config[TRAINER]["regularization_type"] = regularizer

        # run experiment
        _, _, _, _, output_dir = experiment_cli(
            training_set=generated_data.train_df,
            validation_set=generated_data.validation_df,
            test_set=generated_data.test_df,
            output_directory=str(results_dir),
            config=config,
            experiment_name="regularization",
            model_name=str(regularizer),
            skip_save_processed_input=True,
            skip_save_progress=True,
            skip_save_unprocessed_output=True,
            skip_save_model=True,
            skip_save_log=True,
        )

        # test existence of required files
        train_stats_fp = os.path.join(output_dir, "training_statistics.json")
        metadata_fp = os.path.join(output_dir, DESCRIPTION_FILE_NAME)
        assert os.path.isfile(train_stats_fp)
        assert os.path.isfile(metadata_fp)

        # retrieve results so we can compare training loss with regularization
        with open(train_stats_fp) as f:
            train_stats = json.load(f)

        # retrieve training losses for all epochs
        train_losses = train_stats[TRAINING]["combined"]["loss"]
        regularization_losses.append(train_losses[0])

    # create a set of losses
    regularization_losses_set = set(regularization_losses)

    # ensure all losses obtained with the different methods are different
    assert len(regularization_losses) == len(regularization_losses_set)


# test cache checksum function
def test_cache_checksum(csv_filename, tmp_path):
    # setup for training
    input_features = [category_feature(encoder={"vocab_size": 5})]
    output_features = [category_feature(decoder={"vocab_size": 2}, top_k=2)]

    source_dataset = os.path.join(tmp_path, csv_filename)
    source_dataset = generate_data(input_features, output_features, source_dataset)

    config = {
        INPUT_FEATURES: input_features,
        OUTPUT_FEATURES: output_features,
        DEFAULTS: {CATEGORY: {PREPROCESSING: {"fill_value": "<UNKNOWN>"}}},
        TRAINER: {EPOCHS: 2, BATCH_SIZE: 128},
    }

    backend = LocalTestBackend()
    cache_fname = replace_file_extension(source_dataset, TRAINING_PREPROC_FILE_NAME)

    # conduct initial training
    output_directory = os.path.join(tmp_path, "results")
    model = LudwigModel(config, backend=backend)
    model.train(dataset=source_dataset, output_directory=output_directory)
    first_training_timestamp = os.path.getmtime(cache_fname)

    # conduct second training, should not force recreating hdf5
    model = LudwigModel(config, backend=backend)
    model.train(dataset=source_dataset, output_directory=output_directory)
    current_training_timestamp = os.path.getmtime(cache_fname)

    # time stamps should be the same
    assert first_training_timestamp == current_training_timestamp

    # force recreating cache file by changing checksum by updating defaults
    prior_training_timestamp = current_training_timestamp
    config[DEFAULTS][CATEGORY][PREPROCESSING]["fill_value"] = "<EMPTY>"
    model = LudwigModel(config, backend=backend)
    model.train(dataset=source_dataset, output_directory=output_directory)
    current_training_timestamp = os.path.getmtime(cache_fname)

    # timestamp should differ
    assert prior_training_timestamp < current_training_timestamp

    # force recreating cache by updating modification time of source dataset
    prior_training_timestamp = current_training_timestamp
    os.utime(source_dataset)
    model = LudwigModel(config, backend=backend)
    model.train(dataset=source_dataset, output_directory=output_directory)
    current_training_timestamp = os.path.getmtime(cache_fname)

    # timestamps should be different
    assert prior_training_timestamp < current_training_timestamp

    # force change in feature preprocessing
    prior_training_timestamp = current_training_timestamp
    input_features = config[INPUT_FEATURES].copy()
    input_features[0][PREPROCESSING] = {"lowercase": True}
    config[INPUT_FEATURES] = input_features
    model = LudwigModel(config, backend=backend)
    model.train(dataset=source_dataset, output_directory=output_directory)
    current_training_timestamp = os.path.getmtime(cache_fname)

    # timestamps should be different
    assert prior_training_timestamp < current_training_timestamp

    # force change in features names (and properties)
    prior_training_timestamp = current_training_timestamp
    input_features = [category_feature(encoder={"vocab_size": 5}), category_feature()]
    source_dataset = generate_data(input_features, output_features, source_dataset)
    config[INPUT_FEATURES] = input_features
    model = LudwigModel(config, backend=backend)
    model.train(dataset=source_dataset, output_directory=output_directory)
    current_training_timestamp = os.path.getmtime(cache_fname)

    # timestamps should be different
    assert prior_training_timestamp < current_training_timestamp

    # force change in Ludwig version
    prior_training_timestamp = current_training_timestamp
    global_vars.LUDWIG_VERSION = "new_version"
    model = LudwigModel(config, backend=backend)
    model.train(dataset=source_dataset, output_directory=output_directory)
    current_training_timestamp = os.path.getmtime(cache_fname)

    # timestamps should be different
    assert prior_training_timestamp < current_training_timestamp


@pytest.mark.parametrize("transformer_key", list(numeric_transformation_registry.keys()))
def test_numeric_transformer(transformer_key, tmpdir):
    Transformer = get_from_registry(transformer_key, numeric_transformation_registry)
    transformer_name = Transformer().__class__.__name__
    if transformer_name == "Log1pTransformer":
        raw_values = np.random.lognormal(5, 2, size=100)
    else:
        raw_values = np.random.normal(5, 2, size=100)

    backend = LOCAL_BACKEND
    parameters = Transformer.fit_transform_params(raw_values, backend)
    if transformer_name in {"Log1pTransformer", "IdentityTransformer"}:
        # should be empty
        assert not bool(parameters)
    else:
        # should not be empty
        assert bool(parameters)

    # instantiate numeric transformer
    numeric_transfomer = Transformer(**parameters)

    # transform values
    transformed_values = numeric_transfomer.transform(raw_values)

    # inverse transform the prior transformed values
    reconstructed_values = numeric_transfomer.inverse_transform(transformed_values)

    # should now match
    assert np.allclose(raw_values, reconstructed_values)

    # now test numeric transformer with output feature
    df = pd.DataFrame(np.array([raw_values, raw_values]).T, columns=["x", "y"])
    config = {
        "input_features": [{"name": "x", "type": "number"}],
        "output_features": [{"name": "y", "type": "number", "preprocessing": {"normalization": transformer_key}}],
        "combiner": {
            "type": "concat",
        },
        TRAINER: {
            "epochs": 2,
            "batch_size": 16,
        },
    }

    args = {
        "config": config,
        "skip_save_processed_input": True,
        "output_directory": os.path.join(tmpdir, "results"),
        "logging_level": logging.WARN,
    }

    # ensure no exceptions are raised
    experiment_cli(dataset=df, **args)



#! /usr/bin/env python
# Copyright (c) 2019 Uber Technologies, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import numpy as np

from ludwig.data.batcher.base import Batcher


class BucketedBatcher(Batcher):
    def __init__(
        self,
        dataset,
        bucketing_field,
        batch_size=128,
        buckets=10,
        should_shuffle=True,
        ignore_last=False,
        should_trim=False,
        trim_side="right",
    ):
        self.should_shuffle = should_shuffle
        self.bucketing_field = bucketing_field
        self.should_trim = should_trim
        self.trim_side = trim_side

        # store our dataset as well
        self.dataset = dataset

        field = dataset.get_dataset()[bucketing_field]
        field_lengths = np.apply_along_axis(lambda x: np.sign(x).sum(), 1, field)
        sorted_idcs = np.argsort(field_lengths)
        self.buckets_idcs = []
        datapoints_per_bucket = len(field) // buckets
        for b in range(buckets):
            start = datapoints_per_bucket * b
            end = datapoints_per_bucket * (b + 1) if b < buckets - 1 else len(sorted_idcs)
            self.buckets_idcs.append(sorted_idcs[start:end])

        if should_shuffle:
            self.shuffle(self.buckets_idcs)

        self.ignore_last = ignore_last
        self.batch_size = batch_size
        self.total_size = min(map(len, dataset.get_dataset().values()))
        self.bucket_sizes = np.array([x for x in map(len, self.buckets_idcs)])
        self.steps_per_epoch = self._compute_steps_per_epoch()
        self.indices = np.array([0] * buckets)
        self.step = 0
        self.epoch = 0

    def shuffle(self, buckets_idcs):
        for i in range(len(buckets_idcs)):
            np.random.shuffle(buckets_idcs[i])

    def next_batch(self):
        if self.last_batch():
            if self.should_shuffle:
                self.shuffle(self.buckets_idcs)
            self.set_epoch(self.epoch + 1)

        if self.ignore_last:
            idcs_below_size = self.indices + self.batch_size < self.bucket_sizes
        else:
            idcs_below_size = self.indices < self.bucket_sizes
        i = np.random.choice(np.arange(0, len(self.buckets_idcs))[idcs_below_size])

        selected_bucket = self.buckets_idcs[i]
        selected_idcs = selected_bucket[self.indices[i] : self.indices[i] + self.batch_size]

        sub_batch = {}
        for key in self.dataset.get_dataset():
            if key == self.bucketing_field and self.should_trim:
                selected_samples = self.dataset.get(key, selected_idcs)
                max_length = np.sign(selected_samples).sum(axis=1).max()
                if self.trim_side == "right":
                    sub_batch[key] = selected_samples[:, :max_length]
                elif self.trim_side == "left":
                    sub_batch[key] = selected_samples[:, -max_length:]
                else:
                    raise ValueError("Invalid trim side:", self.trim_side)

            else:
                sub_batch[key] = self.dataset.get(key, selected_idcs)

        self.indices[i] += self.batch_size
        self.step += 1
        return sub_batch

    def last_batch(self):
        return not np.any(self.indices < self.bucket_sizes) or (
            self.ignore_last and not np.any(self.indices + self.batch_size < self.bucket_sizes)
        )

    def set_epoch(self, epoch, batch_size):
        self.indices = np.array([0] * len(self.buckets_idcs))
        self.step = 0
        self.epoch = epoch
        self.batch_size = batch_size
        self.steps_per_epoch = self._compute_steps_per_epoch()

    def _compute_steps_per_epoch(self) -> int:
        return int(np.sum(np.ceil(self.bucket_sizes / self.batch_size)).item())


# dynamic_length_encoders = {
#     'rnn',
#     'embed'
# }
#
# todo future: reintroduce the bucketed batcher
# def initialize_batcher(dataset, batch_size=128, bucketing_field=None,
#                        input_features=None, preprocessing=None,
#                        should_shuffle=True, ignore_last=False, distributed=None):
#     if distributed:
#         batcher = DistributedBatcher(
#             dataset,
#             distributed.rank(),
#             distributed,
#             batch_size,
#             should_shuffle=should_shuffle,
#             ignore_last=ignore_last
#         )
#     elif bucketing_field is not None:
#         bucketing_feature = [
#             feature for feature in input_features if
#             feature[NAME] == bucketing_field
#         ]
#         if not bucketing_feature:
#             raise ValueError(
#                 'Bucketing field {} not present in input features'.format(
#                     bucketing_field
#                 )
#             )
#         else:
#             bucketing_feature = bucketing_feature[0]
#         should_trim = bucketing_feature[
#                           'encoder'] in dynamic_length_encoders
#         if 'preprocessing' in bucketing_feature:
#             trim_side = bucketing_feature['preprocessing']['padding']
#         else:
#             trim_side = preprocessing[bucketing_feature[TYPE]]['padding']
#
#         batcher = BucketedBatcher(
#             dataset,
#             bucketing_field=bucketing_field,
#             batch_size=batch_size,
#             buckets=10,
#             ignore_last=ignore_last,
#             should_shuffle=should_shuffle,
#             should_trim=should_trim,
#             trim_side=trim_side
#         )
#     else:
#         batcher = Batcher(
#             dataset,
#             batch_size,
#             should_shuffle=should_shuffle,
#             ignore_last=ignore_last
#         )
#     return batcher



import copy
import math
from typing import Any, Dict

import pytest

from ludwig.constants import (
    BATCH_SIZE,
    BFILL,
    CLASS_WEIGHTS,
    DEFAULTS,
    EVAL_BATCH_SIZE,
    EXECUTOR,
    HYPEROPT,
    INPUT_FEATURES,
    LEARNING_RATE_SCHEDULER,
    LOSS,
    NUMBER,
    OUTPUT_FEATURES,
    PREPROCESSING,
    SCHEDULER,
    SPLIT,
    TRAINER,
    TYPE,
)
from ludwig.schema.model_config import ModelConfig
from ludwig.schema.trainer import ECDTrainerConfig
from ludwig.utils.backward_compatibility import (
    _update_backend_cache_credentials,
    _upgrade_encoder_decoder_params,
    _upgrade_feature,
    _upgrade_preprocessing_split,
    upgrade_config_dict_to_latest_version,
    upgrade_missing_value_strategy,
    upgrade_model_progress,
)
from ludwig.utils.trainer_utils import TrainerMetric


def test_preprocessing_backward_compatibility():
    # From v0.5.3.
    preprocessing_config = {
        "force_split": False,
        "split_probabilities": [0.7, 0.1, 0.2],
        "stratify": None,
    }

    _upgrade_preprocessing_split(preprocessing_config)

    assert preprocessing_config == {
        "split": {"probabilities": [0.7, 0.1, 0.2], "type": "random"},
    }


def test_audio_feature_backward_compatibility():
    # From v0.5.3.

    audio_feature_preprocessing_config = {
        "name": "audio_feature",
        "type": "audio",
        "preprocessing": {
            "audio_file_length_limit_in_s": 7.5,
            "missing_value_strategy": BFILL,
            "in_memory": True,
            "padding_value": 0,
            "norm": None,
            "audio_feature": {
                "type": "fbank",
                "window_length_in_s": 0.04,
                "window_shift_in_s": 0.02,
                "num_fft_points": None,
                "window_type": "hamming",
                "num_filter_bands": 80,
            },
        },
    }

    global_preprocessing_config = {
        "audio": {
            "audio_file_length_limit_in_s": 7.5,
            "missing_value_strategy": BFILL,
            "in_memory": True,
            "padding_value": 0,
            "norm": None,
            "audio_feature": {
                "type": "fbank",
                "window_length_in_s": 0.04,
                "window_shift_in_s": 0.02,
                "num_fft_points": None,
                "window_type": "hamming",
                "num_filter_bands": 80,
            },
        },
    }

    _upgrade_feature(audio_feature_preprocessing_config)
    _upgrade_preprocessing_split(global_preprocessing_config)

    assert global_preprocessing_config == {
        "audio": {
            "audio_file_length_limit_in_s": 7.5,
            "missing_value_strategy": BFILL,
            "in_memory": True,
            "padding_value": 0,
            "norm": None,
            "type": "fbank",
            "window_length_in_s": 0.04,
            "window_shift_in_s": 0.02,
            "num_fft_points": None,
            "window_type": "hamming",
            "num_filter_bands": 80,
        }
    }

    assert audio_feature_preprocessing_config == {
        "name": "audio_feature",
        "type": "audio",
        "preprocessing": {
            "audio_file_length_limit_in_s": 7.5,
            "missing_value_strategy": BFILL,
            "in_memory": True,
            "padding_value": 0,
            "norm": None,
            "type": "fbank",
            "window_length_in_s": 0.04,
            "window_shift_in_s": 0.02,
            "num_fft_points": None,
            "window_type": "hamming",
            "num_filter_bands": 80,
        },
    }


def test_encoder_decoder_backwards_compatibility():
    old_config = {
        "input_features": [
            {
                "name": "text_feature",
                "type": "text",
                "preprocessing": {
                    "missing_value_strategy": "drop_row",
                },
                "encoder": "rnn",
                "bidirectional": True,
                "representation": "dense",
                "num_layers": 2,
            },
            {
                "name": "image_feature_1",
                "type": "image",
                "preprocessing": {
                    "height": 7.5,
                    "width": 7.5,
                    "num_channels": 4,
                },
                "encoder": "resnet",
                "num_channels": 4,
                "dropout": 0.1,
                "resnet_size": 100,
            },
            {
                "name": "image_feature_2",
                "type": "image",
                "tied": "image_feature_1",
                "preprocessing": {
                    "height": 7.5,
                    "width": 7.5,
                    "num_channels": 4,
                },
                "encoder": "resnet",
            },
        ],
        "output_features": [
            {
                "name": "category_feature",
                "type": "category",
                "top_k": 3,
                "preprocessing": {
                    "missing_value_strategy": BFILL,
                },
                "decoder": "classifier",
                "num_classes": 10,
                "use_bias": False,
            },
            {
                "name": "binary_feature",
                "type": "binary",
                "dependencies": ["category_feature"],
                "loss": {
                    "type": "cross_entropy",
                },
                "reduce_dependencies": "mean",
                "decoder": "regressor",
                "use_bias": True,
                "bias_initializer": "constant",
            },
            {
                "name": "vector_feature",
                "type": "vector",
                "decoder": "projector",
                "num_fc_layers": 5,
                "output_size": 128,
                "activation": "tanh",
                "dropout": 0.1,
            },
        ],
    }

    for feature in old_config[INPUT_FEATURES]:
        _upgrade_encoder_decoder_params(feature, True)

    for feature in old_config[OUTPUT_FEATURES]:
        _upgrade_encoder_decoder_params(feature, False)

    assert old_config == {
        "input_features": [
            {
                "name": "text_feature",
                "type": "text",
                "preprocessing": {
                    "missing_value_strategy": "drop_row",
                },
                "encoder": {
                    "type": "rnn",
                    "bidirectional": True,
                    "representation": "dense",
                    "num_layers": 2,
                },
            },
            {
                "name": "image_feature_1",
                "type": "image",
                "preprocessing": {
                    "height": 7.5,
                    "width": 7.5,
                    "num_channels": 4,
                },
                "encoder": {
                    "type": "resnet",
                    "num_channels": 4,
                    "dropout": 0.1,
                    "resnet_size": 100,
                },
            },
            {
                "name": "image_feature_2",
                "type": "image",
                "tied": "image_feature_1",
                "preprocessing": {
                    "height": 7.5,
                    "width": 7.5,
                    "num_channels": 4,
                },
                "encoder": {"type": "resnet"},
            },
        ],
        "output_features": [
            {
                "name": "category_feature",
                "type": "category",
                "num_classes": 10,
                "top_k": 3,
                "preprocessing": {
                    "missing_value_strategy": BFILL,
                },
                "decoder": {
                    "type": "classifier",
                    "fc_use_bias": False,
                    "use_bias": False,
                },
            },
            {
                "name": "binary_feature",
                "type": "binary",
                "dependencies": ["category_feature"],
                "loss": {
                    "type": "cross_entropy",
                },
                "reduce_dependencies": "mean",
                "decoder": {
                    "type": "regressor",
                    "fc_use_bias": True,
                    "fc_bias_initializer": "constant",
                    "bias_initializer": "constant",
                    "use_bias": True,
                },
            },
            {
                "name": "vector_feature",
                "type": "vector",
                "decoder": {
                    "type": "projector",
                    "num_fc_layers": 5,
                    "fc_output_size": 128,
                    "fc_activation": "tanh",
                    "fc_dropout": 0.1,
                    "output_size": 128,
                    "activation": "tanh",
                    "dropout": 0.1,
                },
            },
        ],
    }


def test_deprecated_field_aliases():
    config = {
        "ludwig_version": "0.4",
        INPUT_FEATURES: [{"name": "num_in", "type": "numerical"}],
        OUTPUT_FEATURES: [{"name": "num_out", "type": "numerical"}],
        HYPEROPT: {
            "parameters": {
                "training.learning_rate": {
                    "space": "loguniform",
                    "lower": 0.001,
                    "upper": 0.1,
                },
            },
            "goal": "minimize",
            "sampler": {"type": "grid", "num_samples": 2, "scheduler": {"type": "fifo"}},
            "executor": {
                "type": "grid",
                "search_alg": "bohb",
            },
        },
        PREPROCESSING: {
            "numerical": {
                "fill_value": 2,
                "missing_value_strategy": "fill_with_const",
            },
        },
        "training": {
            "epochs": 2,
            "eval_batch_size": 0,
            "reduce_learning_rate_on_plateau": 2,
            "reduce_learning_rate_on_plateau_patience": 5,
            "decay": True,
            "learning_rate_warmup_epochs": 2,
        },
    }

    updated_config = upgrade_config_dict_to_latest_version(config)

    assert updated_config["input_features"][0][TYPE] == NUMBER
    assert updated_config["output_features"][0][TYPE] == NUMBER

    # "numerical" preprocssing directive should be translated to "number" and moved into the defaults section.
    assert PREPROCESSING not in updated_config
    assert updated_config[DEFAULTS][NUMBER][PREPROCESSING]["fill_value"] == 2

    assert "training" not in updated_config
    assert updated_config[TRAINER]["epochs"] == 2
    assert updated_config[TRAINER][EVAL_BATCH_SIZE] is None

    assert LEARNING_RATE_SCHEDULER in updated_config[TRAINER]
    assert updated_config[TRAINER][LEARNING_RATE_SCHEDULER]["reduce_on_plateau"] == 2
    assert updated_config[TRAINER][LEARNING_RATE_SCHEDULER]["reduce_on_plateau_patience"] == 5
    assert updated_config[TRAINER][LEARNING_RATE_SCHEDULER]["decay"] == "exponential"
    assert updated_config[TRAINER][LEARNING_RATE_SCHEDULER]["warmup_evaluations"] == 2

    hparams = updated_config[HYPEROPT]["parameters"]
    assert "training.learning_rate" not in hparams
    assert "trainer.learning_rate" in hparams

    assert "sampler" not in updated_config[HYPEROPT]

    assert updated_config[HYPEROPT]["executor"]["type"] == "ray"
    assert "num_samples" in updated_config[HYPEROPT]["executor"]
    assert "scheduler" in updated_config[HYPEROPT]["executor"]

    ModelConfig.from_dict(updated_config)


@pytest.mark.parametrize("force_split", [None, False, True])
@pytest.mark.parametrize("stratify", [None, "cat_in"])
def test_deprecated_split_aliases(stratify, force_split):
    split_probabilities = [0.6, 0.2, 0.2]
    config = {
        "ludwig_version": "0.4",
        INPUT_FEATURES: [{"name": "num_in", "type": "number"}, {"name": "cat_in", "type": "category"}],
        OUTPUT_FEATURES: [{"name": "num_out", "type": "number"}],
        PREPROCESSING: {
            "force_split": force_split,
            "split_probabilities": split_probabilities,
            "stratify": stratify,
        },
    }

    updated_config = upgrade_config_dict_to_latest_version(config)

    assert "force_split" not in updated_config[PREPROCESSING]
    assert "split_probabilities" not in updated_config[PREPROCESSING]
    assert "stratify" not in updated_config[PREPROCESSING]

    assert SPLIT in updated_config[PREPROCESSING]
    split = updated_config[PREPROCESSING][SPLIT]

    assert split["probabilities"] == split_probabilities
    if stratify is None:
        if force_split:
            assert split.get(TYPE) == "random"
    else:
        assert split.get(TYPE) == "stratify"
        assert split.get("column") == stratify


@pytest.mark.parametrize("use_scheduler", [True, False])
def test_deprecated_hyperopt_sampler_early_stopping(use_scheduler):
    sampler = {
        "type": "ray",
        "num_samples": 2,
    }

    if use_scheduler:
        sampler[SCHEDULER] = {
            "type": "async_hyperband",
            "max_t": 200,
            "time_attr": "time_total_s",
            "grace_period": 72,
            "reduction_factor": 5,
        }

    config = {
        INPUT_FEATURES: [
            {
                "type": "category",
                "name": "cat_input_feature",
            },
        ],
        OUTPUT_FEATURES: [
            {
                "type": "number",
                "name": "num_output_feature",
            },
        ],
        "hyperopt": {
            "search_alg": {
                "type": "hyperopt",
                "random_state_seed": 42,
            },
            "executor": {
                "type": "ray",
                "time_budget_s": 200,
                "cpu_resources_per_trial": 1,
            },
            "sampler": sampler,
            "parameters": {
                "trainer.batch_size": {
                    "space": "choice",
                    "categories": [64, 128, 256],
                },
                "trainer.learning_rate": {
                    "space": "loguniform",
                    "lower": 0.001,
                    "upper": 0.1,
                },
            },
        },
    }

    updated_config = upgrade_config_dict_to_latest_version(config)
    if use_scheduler:
        assert SCHEDULER in updated_config[HYPEROPT][EXECUTOR]

    merged_config = ModelConfig.from_dict(updated_config).to_dict()

    # When a scheulder is provided, early stopping in the rendered config needs to be disabled to allow the
    # hyperopt scheduler to manage trial lifecycle.
    expected_early_stop = -1 if use_scheduler else ECDTrainerConfig().early_stop
    assert merged_config[TRAINER]["early_stop"] == expected_early_stop


def test_validate_old_model_config():
    old_valid_config = {
        "input_features": [
            {"name": "feature_1", "type": "category"},
            {"name": "Sex", "type": "category", "encoder": "dense"},
        ],
        "output_features": [
            {"name": "Survived", "type": "category"},
        ],
    }

    old_invalid_config = {
        "input_features": [
            {"name": "feature_1", "type": "category"},
            {"name": "Sex", "type": "category", "encoder": "fake_encoder"},
        ],
        "output_features": [
            {"name": "Survived", "type": "category"},
        ],
    }

    ModelConfig.from_dict(old_valid_config)

    with pytest.raises(Exception):
        ModelConfig.from_dict(old_invalid_config)


@pytest.mark.parametrize("missing_value_strategy", ["backfill", "pad"])
def test_update_missing_value_strategy(missing_value_strategy: str):
    old_valid_config = {
        "input_features": [
            {
                "name": "input_feature_1",
                "type": "category",
                "preprocessing": {"missing_value_strategy": missing_value_strategy},
            }
        ],
        "output_features": [
            {"name": "output_feature_1", "type": "category"},
        ],
    }

    updated_config = upgrade_missing_value_strategy(old_valid_config)

    expected_config = copy.deepcopy(old_valid_config)
    if missing_value_strategy == "backfill":
        expected_config["input_features"][0]["preprocessing"]["missing_value_strategy"] == "bfill"
    else:
        expected_config["input_features"][0]["preprocessing"]["missing_value_strategy"] == "ffill"

    assert updated_config == expected_config


def test_update_increase_batch_size_on_plateau_max():
    old_valid_config = {
        "input_features": [{"name": "input_feature_1", "type": "category"}],
        "output_features": [{"name": "output_feature_1", "type": "category"}],
        "trainer": {
            "increase_batch_size_on_plateau_max": 256,
        },
    }

    updated_config = upgrade_config_dict_to_latest_version(old_valid_config)
    del updated_config["ludwig_version"]

    expected_config = copy.deepcopy(old_valid_config)
    del expected_config["trainer"]["increase_batch_size_on_plateau_max"]
    expected_config["trainer"]["max_batch_size"] = 256

    assert updated_config == expected_config


def test_old_class_weights_default():
    old_config = {
        "input_features": [
            {
                "name": "input_feature_1",
                "type": "category",
            }
        ],
        "output_features": [
            {"name": "output_feature_1", "type": "category", "loss": {"class_weights": 1}},
        ],
    }

    new_config = {
        "input_features": [
            {
                "name": "input_feature_1",
                "type": "category",
            }
        ],
        "output_features": [
            {"name": "output_feature_1", "type": "category", "loss": {"class_weights": None}},
        ],
    }

    upgraded_config = upgrade_config_dict_to_latest_version(old_config)
    del upgraded_config["ludwig_version"]
    assert new_config == upgraded_config

    old_config[OUTPUT_FEATURES][0][LOSS][CLASS_WEIGHTS] = [0.5, 0.8, 1]
    new_config[OUTPUT_FEATURES][0][LOSS][CLASS_WEIGHTS] = [0.5, 0.8, 1]

    upgraded_config = upgrade_config_dict_to_latest_version(old_config)
    del upgraded_config["ludwig_version"]
    assert new_config == upgraded_config


def test_upgrade_model_progress():
    old_model_progress = {
        "batch_size": 64,
        "best_eval_metric": 0.5,
        "best_increase_batch_size_eval_metric": math.inf,
        "best_reduce_learning_rate_eval_metric": math.inf,
        "epoch": 2,
        "last_improvement": 1,
        "last_improvement_epoch": 1,
        "best_eval_metric_epoch": 1,
        "last_increase_batch_size": 0,
        "last_increase_batch_size_epoch": 0,
        "last_increase_batch_size_eval_metric_improvement": 0,
        "last_learning_rate_reduction": 0,
        "last_learning_rate_reduction_epoch": 0,
        "last_reduce_learning_rate_eval_metric_improvement": 0,
        "learning_rate": 0.001,
        "num_increases_batch_size": 0,
        "num_reductions_learning_rate": 0,
        "steps": 224,
        "test_metrics": {
            "combined": {"loss": [0.59, 0.56]},
            "delinquent": {
                "accuracy": [0.77, 0.78],
            },
        },
        "train_metrics": {"combined": {"loss": [0.58, 0.55]}, "delinquent": {"roc_auc": [0.53, 0.54]}},
        "vali_metrics": {"combined": {"loss": [0.59, 0.60]}, "delinquent": {"roc_auc": [0.53, 0.44]}},
    }

    new_model_progress = upgrade_model_progress(old_model_progress)

    assert new_model_progress == {
        "batch_size": 64,
        "best_eval_metric_value": 0.5,
        "best_increase_batch_size_eval_metric": float("inf"),
        "epoch": 2,
        "last_improvement_steps": 64,
        "best_eval_metric_steps": 0,
        "best_eval_metric_epoch": 1,
        "last_increase_batch_size": 0,
        "last_increase_batch_size_eval_metric_improvement": 0,
        "last_learning_rate_reduction": 0,
        "learning_rate": 0.001,
        "num_increases_batch_size": 0,
        "num_reductions_learning_rate": 0,
        "steps": 224,
        "test_metrics": {
            "combined": {
                "loss": [TrainerMetric(epoch=1, step=64, value=0.59), TrainerMetric(epoch=2, step=128, value=0.56)]
            },
            "delinquent": {
                "accuracy": [TrainerMetric(epoch=1, step=64, value=0.77), TrainerMetric(epoch=2, step=128, value=0.78)]
            },
        },
        "train_metrics": {
            "combined": {
                "loss": [TrainerMetric(epoch=1, step=64, value=0.58), TrainerMetric(epoch=2, step=128, value=0.55)]
            },
            "delinquent": {
                "roc_auc": [TrainerMetric(epoch=1, step=64, value=0.53), TrainerMetric(epoch=2, step=128, value=0.54)]
            },
        },
        "last_learning_rate_reduction_steps": 0,
        "last_increase_batch_size_steps": 0,
        "validation_metrics": {
            "combined": {
                "loss": [TrainerMetric(epoch=1, step=64, value=0.59), TrainerMetric(epoch=2, step=128, value=0.6)]
            },
            "delinquent": {
                "roc_auc": [TrainerMetric(epoch=1, step=64, value=0.53), TrainerMetric(epoch=2, step=128, value=0.44)]
            },
        },
        "tune_checkpoint_num": 0,
        "checkpoint_number": 0,
        "best_eval_metric_checkpoint_number": 0,
        "best_eval_train_metrics": {},
        "best_eval_validation_metrics": {},
        "best_eval_test_metrics": {},
    }

    # Verify that we don't make changes to already-valid model progress dicts.
    # To do so, we modify the batch size value and re-run the upgrade on the otherwise-valid `new_model_progress` dict.
    new_model_progress["batch_size"] = 1
    unchanged_model_progress = upgrade_model_progress(new_model_progress)
    assert unchanged_model_progress == new_model_progress


def test_upgrade_model_progress_already_valid():
    # Verify that we don't make changes to already-valid model progress dicts.
    valid_model_progress = {
        BATCH_SIZE: 128,
        "best_eval_metric_checkpoint_number": 7,
        "best_eval_metric_epoch": 6,
        "best_eval_metric_steps": 35,
        "best_eval_metric_value": 0.719,
        "best_eval_test_metrics": {
            "Survived": {"accuracy": 0.634, "loss": 3.820, "roc_auc": 0.598},
            "combined": {"loss": 3.820},
        },
        "best_eval_train_metrics": {
            "Survived": {"accuracy": 0.682, "loss": 4.006, "roc_auc": 0.634},
            "combined": {"loss": 4.006},
        },
        "best_eval_validation_metrics": {
            "Survived": {"accuracy": 0.719, "loss": 4.396, "roc_auc": 0.667},
            "combined": {"loss": 4.396},
        },
        "best_increase_batch_size_eval_metric": float("inf"),
        "checkpoint_number": 12,
        "epoch": 12,
        "last_increase_batch_size": 0,
        "last_increase_batch_size_eval_metric_improvement": 0,
        "last_increase_batch_size_steps": 0,
        "last_learning_rate_reduction": 0,
        "last_learning_rate_reduction_steps": 0,
        "learning_rate": 0.001,
        "num_increases_batch_size": 0,
        "num_reductions_learning_rate": 0,
        "steps": 60,
        "test_metrics": {
            "Survived": {
                "accuracy": [
                    [0, 5, 0.651],
                    [1, 10, 0.651],
                ],
                "loss": [
                    [0, 5, 4.130],
                    [1, 10, 4.074],
                ],
                "roc_auc": [
                    [0, 5, 0.574],
                    [1, 10, 0.595],
                ],
            },
            "combined": {
                "loss": [
                    [0, 5, 4.130],
                    [1, 10, 4.074],
                ]
            },
        },
        "train_metrics": {
            "Survived": {
                "accuracy": [
                    [0, 5, 0.6875],
                    [1, 10, 0.6875],
                ],
                "loss": [
                    [0, 5, 4.417],
                    [1, 10, 4.344],
                ],
                "roc_auc": [
                    [0, 5, 0.628],
                    [1, 10, 0.629],
                ],
            },
            "combined": {
                "loss": [
                    [0, 5, 4.417],
                    [1, 10, 4.344],
                ]
            },
        },
        "tune_checkpoint_num": 0,
        "validation_metrics": {
            "Survived": {
                "accuracy": [
                    [0, 5, 0.696],
                    [1, 10, 0.696],
                ],
                "loss": [
                    [0, 5, 4.494],
                    [1, 10, 4.473],
                ],
                "roc_auc": [
                    [0, 5, 0.675],
                    [1, 10, 0.671],
                ],
            },
            "combined": {
                "loss": [
                    [0, 5, 4.494],
                    [1, 10, 4.473],
                ]
            },
        },
    }

    unchanged_model_progress = upgrade_model_progress(valid_model_progress)
    assert unchanged_model_progress == valid_model_progress


def test_cache_credentials_backward_compatibility():
    # From v0.6.3.
    creds = {"s3": {"client_kwargs": {}}}
    backend = {"type": "local", "cache_dir": "/foo/bar", "cache_credentials": creds}

    _update_backend_cache_credentials(backend)

    assert backend == {"type": "local", "cache_dir": "/foo/bar", "credentials": {"cache": creds}}


@pytest.mark.parametrize(
    "encoder,upgraded_type",
    [
        ({"type": "resnet"}, "resnet"),
        ({"type": "vit"}, "vit"),
        ({"type": "resnet", "resnet_size": 50}, "_resnet_legacy"),
        ({"type": "vit", "num_hidden_layers": 12}, "_vit_legacy"),
    ],
    ids=["resnet", "vit", "resnet_legacy", "vit_legacy"],
)
def test_legacy_image_encoders(encoder: Dict[str, Any], upgraded_type: str):
    config = {
        "input_features": [{"name": "image1", "type": "image", "encoder": encoder}],
        "output_features": [{"name": "binary1", "type": "binary"}],
    }

    updated_config = upgrade_config_dict_to_latest_version(config)

    expected_encoder = {
        **encoder,
        **{"type": upgraded_type},
    }
    assert updated_config["input_features"][0]["encoder"] == expected_encoder


def test_load_config_missing_hyperopt():
    old_valid_config = {
        "input_features": [
            {"name": "feature_1", "type": "category"},
            {"name": "Sex", "type": "category", "encoder": "dense"},
        ],
        "output_features": [
            {"name": "Survived", "type": "category"},
        ],
        "combiner": {"type": "concat"},
        "trainer": {},
        "hyperopt": {},
    }

    config_obj = ModelConfig.from_dict(old_valid_config)
    assert config_obj.hyperopt is None
    assert config_obj.to_dict()[HYPEROPT] is None
